import numpy as np
from scipy.linalg import circulant
from scipy.optimize import curve_fit, brentq
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter
import scipy.ndimage as nd


def gradient(f, x=None, dx=1, axis=-1):
    """
    Return the gradient of 1 or 2-dimensional array.
    The gradient is computed using central differences in the interior
    and first differences at the boundaries.

    Irregular sampling is supported (it isn't supported by np.gradient)

    Parameters
    ----------
    f : 1d or 2d numpy array
        Input array.
    x : array_like, optional
       Points where the function f is evaluated. It must be of the same
       length as ``f.shape[axis]``.
       If None, regular sampling is assumed (see dx)
    dx : float, optional
       If `x` is None, spacing given by `dx` is assumed. Default is 1.
    axis : int, optional
       The axis along which the difference is taken.

    Returns
    -------
    out : array_like
        Returns the gradient along the given axis.

    Notes
    -----
    To-Do: implement smooth noise-robust differentiators for use on experimental data.
    http://www.holoborodko.com/pavel/numerical-methods/numerical-derivative/smooth-low-noise-differentiators/
    """

    if x is None:
        x = np.arange(f.shape[axis]) * dx
    else:
        assert x.shape[0] == f.shape[axis]
    I = np.zeros(f.shape[axis])
    I[:2] = np.array([0, -1])
    I[-1] = 1
    I = circulant(I)
    I[0, 0] = -1
    I[-1, -1] = 1
    I[0, -1] = 0
    I[-1, 0] = 0
    H = np.zeros((f.shape[axis], 1))
    H[1:-1, 0] = x[2:] - x[:-2]
    H[0] = x[1] - x[0]
    H[-1] = x[-1] - x[-2]
    if axis == 0:
        return np.dot(I / H, f)
    else:
        return np.dot(I / H, f.T).T


def gaussian(x, a, mu, sigma, c):
    r"""
    `Gaussian function <https://en.wikipedia.org/wiki/Gaussian_function>`_

    :math:`f(x)=a e^{-(x - \mu)^2 / (2 \sigma^2)} + c`

    Parameters
    ----------
    x : 1D np.array
        coordinate

    a : float
        the height of the curve's peak

    mu : float
        the position of the center of the peak

    sigma : float
        the standard deviation, sometimes called the Gaussian RMS width

    c : float
        non-zero background

    Returns
    -------
    out : 1D np.array
        the Gaussian profile
    """
    return a * np.exp(-((x - mu) ** 2) / 2 / sigma ** 2) + c


def guss_gaussian(x):
    """
    Find a set of better starting parameters for Gaussian function fitting

    Parameters
    ----------
    x : 1D np.array
        1D profile of your data

    Returns
    -------
    out : tuple of float
        estimated value of (a, mu, sigma, c)
    """
    c_guess = (x[0] + x[-1]) / 2
    a_guess = x.max() - c_guess
    mu_guess = x.argmax()
    x_inter = interp1d(range(len(x)), x)

    def _(i):
        return x_inter(i) - a_guess / 2 - c_guess

    try:
        sigma_l_guess = brentq(_, 0, mu_guess)
    except:
        sigma_l_guess = len(x) / 4
    try:
        sigma_r_guess = brentq(_, mu_guess, len(x) - 1)
    except:
        sigma_r_guess = 3 * len(x) / 4
    return a_guess, mu_guess, (sigma_r_guess -
                               sigma_l_guess) / 2.35482, c_guess


def fit_gaussian(x):
    """
    Fit a Gaussian function to x and return its parameters

    Parameters
    ----------
    x : 1D np.array
        1D profile of your data

    Returns
    -------
    out : tuple of float
        (a, mu, sigma, c)
    """
    p, q = curve_fit(gaussian, list(range(x.size)), x, p0=guss_gaussian(x))
    return p



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import time
from warnings import warn
import platform
import os
import sys

from . import basex
from . import dasch
from . import daun
from . import direct
from . import hansenlaw
from . import linbasex
from . import onion_bordas
from . import rbasex
from . import tools

from abel import _deprecated, _deprecate


class Transform(object):
    r"""
    Abel transform image class.

    This class provides whole-image forward and inverse Abel
    transforms, together with preprocessing (centering, symmetrizing)
    and postprocessing (integration) functions.

    Parameters
    ----------

    IM : a N×M numpy array
        This is the image to be transformed

    direction : str
        The type of Abel transform to be performed.

        ``forward``
            A forward Abel transform takes a (2D) slice of a 3D image and
            returns the 2D projection.

        ``inverse`` (default)
            An inverse Abel transform takes a 2D projection and reconstructs
            a 2D slice of the 3D image.

    method : str
        specifies which numerical approximation to the Abel transform should be
        employed (see below). The options are

        ``basex``
            the Gaussian "basis set expansion" method of Dribinski et al.
            (2002).
        ``daun``
            the deconvolution method with Tikhonov regularization of Daun et
            al. and its extensions.
        ``direct``
            a naive implementation of the analytical formula by Roman Yurchuk.
        ``hansenlaw``
            the recursive algorithm described by Hansen and Law (1985).
        ``linbasex``
            the 1D projections of velocity-mapping images in terms of 1D
            spherical functions by Gerber et al. (2013).
        ``onion_bordas``
            the algorithm of Bordas and co-workers (1996),
            re-implemented by Rallis, Wells and co-workers (2014).
        ``onion_peeling``
            the onion peeling deconvolution as described by Dasch (1992).
        ``rbasex``
            a method similar to pBasex by Garcia et al. (2004) for
            velocity-mapping images, but with analytical basis functions
            developed by Ryazanov (2012).
        ``three_point``
            the three-point transform of Dasch (1992).
        ``two_point``
            the two-point transform of Dasch (1992).

    origin : tuple or str
        Before applying Abel transform, the image is centered around this
        point.

        If a tuple (float, float) is provided, this specifies the image origin
        in the (row, column) format. If a string is provided, an automatic
        centering algorithm is used:

        ``image_center``
            The origin is assumed to be the center of the image.
        ``convolution``
            The origin is found from autoconvolution of image projections
            along each axis.
        ``slice``
            The origin is found by comparing slices in the horizontal and
            vertical directions.
        ``com``
            The origin is calculated as the center of mass.
        ``gaussian``
            The origin is found using a fit to a Gaussian function. This only
            makes sense if your data looks like a Gaussian.
        ``none`` (default)
            No centering is performed. An image with an odd number of columns
            must be provided.

    symmetry_axis : None, int or tuple
        Symmetrize the image about the numpy axis
        0 (vertical), 1 (horizontal), (0, 1) (both axes). Note that the
        Abel transform is always performed around the vertical axis.
        This parameter only affects how the image is modified before
        (and after) applying the Abel transform. For more information,
        see the "Quadrant combining" note below.

    use_quadrants : tuple of 4 booleans
        select quadrants to be used in the analysis: (Q0, Q1, Q2, Q3).
        Quadrants are numbered counter-clockwide from upper right.
        See note below for description of quadrants.
        Default is ``(True, True, True, True)``, which uses all quadrants.

    symmetrize_method : str
        Method used for symmetrizing the image.

        ``average``
            Average the quadrants, in accordance with the **symmetry_axis**.
        ``fourier``
            Axial symmetry implies that the Fourier components of the 2D
            projection should be real. Removing the imaginary components in
            reciprocal space leaves a symmetric projection.

            K. R. Overstreet, P. Zabawa, J. Tallant, A. Schwettmann,
            J. P. Shaffer,
            "Multiple scattering and the density distribution of a Cs MOT",
            `Optics Express 13, 9672–9682 (2005)
            <https://doi.org/10.1364/OPEX.13.009672>`__.

    angular_integration : bool
        Integrate the image over angle to give the radial (speed) intensity
        distribution.

        *Note: in PyAbel ≤0.8.4 the intensity distribution was off by a factor
        of π, please keep this in mind when comparing absolute intensities.*

    transform_options : dict
        Additional arguments passed to the individual transform functions.
        See the documentation for the individual transform method for options.

    center_options : dict
        Additional arguments to be passed to the centering function,
        see :func:`abel.tools.center.center_image()`.

    angular_integration_options : dict
        Additional arguments passed to the angular integration functions,
        see :func:`abel.tools.vmi.angular_integration_3D()`.

    recast_as_float64 : bool
        determines whether the input image should be recast to
        ``float64``. Many images are imported in other formats (such as
        ``uint8`` or ``uint16``), and this does not always play well with the
        transorm algorithms. This should probably always be set to ``True``
        (default).

    verbose : bool
        determines whether non-critical output should be printed.


    .. note::
        Quadrant combining:
        The quadrants can be combined (averaged) using the ``use_quadrants``
        keyword in order to provide better data quality.

        The quadrants are numbered starting from Q0 in the upper right and
        proceeding counter-clockwise::

            +--------+--------+
            | Q1   * | *   Q0 |
            |   *    |    *   |
            |  *     |     *  |                                 AQ1 | AQ0
            +--------o--------+ --([inverse] Abel transform)--> ----o----
            |  *     |     *  |                                 AQ2 | AQ3
            |   *    |    *   |
            | Q2  *  | *   Q3 |          AQi == [inverse] Abel transform
            +--------+--------+                 of quadrant Qi

        Three cases are possible:

        1) symmetry_axis = 0 (vertical)::

            Combine:  Q01 = Q0 + Q1, Q23 = Q2 + Q3
            inverse image   AQ01 | AQ01
                            -----o----- (left and right sides equivalent)
                            AQ23 | AQ23


        2) symmetry_axis = 1 (horizontal)::

            Combine: Q12 = Q1 + Q2, Q03 = Q0 + Q3
            inverse image   AQ12 | AQ03
                            -----o----- (top and bottom equivalent)
                            AQ12 | AQ03


        3) symmetry_axis = (0, 1) (both)::

            Combine: Q = Q0 + Q1 + Q2 + Q3
            inverse image   AQ | AQ
                            ---o---  (all quadrants equivalent)
                            AQ | AQ

    Notes
    -----
    As mentioned above, PyAbel offers several different approximations to the
    the exact Abel transform. All the methods should produce similar
    results, but depending on the level and type of noise found in the image,
    certain methods may perform better than others. Please see the
    :ref:`TransformMethods` section of the documentation for complete
    information.

    The methods marked with a * indicate methods that generate basis sets. The
    first time they are run for a new image size, it takes seconds to minutes
    to generate the basis set. However, this basis set is saved to disk can
    be reloaded, meaning that future transforms are performed much more
    quickly.

    ``basex`` *
        The "basis set exapansion" algorithm describes the data in terms of
        gaussian-like functions, which themselves can be Abel-transformed
        analytically. With the default functions, centered at each pixel,
        this method also does not make any assumption about the
        shape of the data. This method is one of the de-facto standards in
        photoelectron/photoion imaging.

        V. Dribinski, A. Ossadtchi, V. A. Mandelshtam, H. Reisler,
        "Reconstruction of Abel-transformable images: The Gaussian basis-set
        expansion Abel transform method",
        `Rev. Sci. Instrum. 73, 2634–2642 (2002)
        <https://doi.org/10.1063/1.1482156>`__.

    ``daun`` *
        Methods based on onion-peeling deconvolution using Tikhonov
        regularization described in

        K. J. Daun, K. A. Thomson, F. Liu, G. J. Smallwood,
        "Deconvolution of axisymmetric flame properties using Tikhonov
        regularization",
        `Appl. Opt. 45, 4638–4646 (2006)
        <https://doi.org/10.1364/AO.45.004638>`__.

        In addition to the original implicit step-functions basis (“onion
        peeling”) and the derivative regularization, linear and quadratic basis
        functions are implemented, as well as the :math:`L_2`-norm Tikhonov
        regularization (like in ``basex``) and non-negative least-squares
        solution.

    ``direct``
        This method attempts a direct integration of the Abel-transform
        integral. It makes no assumptions about the data (apart from
        cylindrical symmetry), but it typically requires fine sampling to
        converge. Such methods are typically inefficient, but thanks to this
        Cython implementation (by Roman Yurchuk), this "direct" method is
        competitive with the other methods.

    ``hansenlaw``
        This "recursive algorithm" produces reliable results and is quite fast
        (~0.1 s for a 1001×1001 image). It makes no assumptions about the data
        (apart from cylindrical symmetry). It tends to require that the data is
        finely sampled for good convergence.

        E. W. Hansen, P.-L. Law,
        "Recursive methods for computing the Abel transform and its inverse",
        `J. Opt. Soc. Am. A 2, 510–520 (1985)
        <https://doi.org/10.1364/JOSAA.2.000510>`__.

    ``linbasex`` *
        Velocity-mapping images are composed of projected Newton spheres with
        a common centre. The 2D images are usually evaluated by a
        decomposition into base vectors, each representing the 2D projection
        of a set of particles starting from a centre with a specific velocity
        distribution. Lin-BASEX evaluates 1D projections of VM images in terms
        of 1D projections of spherical functions, instead.

        Th. Gerber, Yu. Liu, G. Knopp, P. Hemberger, A. Bodi, P. Radi,
        Ya. Sych,
        "Charged particle velocity map image reconstruction with
        one-dimensional projections of spherical functions",
        `Rev. Sci. Instrum. 84, 033101 (2013)
        <https://doi.org/10.1063/1.4793404>`__.

    ``onion_bordas``
        The onion peeling method, also known as "back projection", originates
        from
        C. Bordas, F. Paulig,
        "Photoelectron imaging spectrometry: Principle and inversion method",
        `Rev. Sci. Instrum. 67, 2257–2268 (1996)
        <https://doi.org/10.1063/1.1147044>`__.

        The algorithm was subsequently coded in MatLab by
        C. E. Rallis, T. G. Burwitz, P. R. Andrews, M. Zohrabi, R. Averin,
        S. De, B. Bergues, B. Jochim, A. V. Voznyuk, N. Gregerson, B. Gaire,
        I. Znakovskaya, J. McKenna, K. D. Carnes, M. F. Kling, I. Ben-Itzhak,
        E. Wells,
        "Incorporating real time velocity map image reconstruction into
        closed-loop coherent control",
        `Rev. Sci. Instrum. 85, 113105 (2014)
        <https://doi.org/10.1063/1.4899267>`__,
        which was used as the basis of this Python port. See `issue #56
        <https://github.com/PyAbel/PyAbel/issues/56>`__.

    ``onion_peeling`` *
        This is one of the most compact and fast algorithms, with the inverse
        Abel transform achieved in one Python code-line, `PR #155
        <https://github.com/PyAbel/PyAbel/pull/155>`__. See also
        ``three_point`` is the onion peeling algorithm as described by Dasch
        (1992), reference below.

    ``rbasex`` *
        The pBasex method by
        G. A. Garcia, L. Nahon, I. Powis,
        “Two-dimensional charged particle image inversion using a polar basis
        function expansion”,
        `Rev. Sci. Instrum. 75, 4989–2996 (2004)
        <http://doi.org/10.1063/1.1807578>`__
        adapts the BASEX ("basis set expansion") method to the specific case of
        velocity-mapping images by using a basis of 2D functions in polar
        coordinates, such that the reconstructed radial distributions are
        obtained directly from the expansion coefficients.

        This method employs the same approach, but uses more convenient basis
        functions, which have analytical Abel transforms separable into radial
        and angular parts, developed in
        M. Ryazanov,
        “Development and implementation of methods for sliced velocity map
        imaging. Studies of overtone-induced dissociation and isomerization
        dynamics of hydroxymethyl radical (CH\ :sub:`2`\ OH and
        CD\ :sub:`2`\ OH)”,
        Ph.D. dissertation, University of Southern California, 2012
        (`ProQuest <https://search.proquest.com/docview/1289069738>`__,
        `USC <https://digitallibrary.usc.edu/asset-management/2A3BF169XWB4>`__).

    ``three_point`` *
        The "Three Point" Abel transform method exploits the observation that
        the value of the Abel inverted data at any radial position r is
        primarily determined from changes in the projection data in the
        neighborhood of r. This method is also very efficient once it has
        generated the basis sets.

        C. J. Dasch,
        "One-dimensional tomography: a comparison of Abel, onion-peeling, and
        filtered backprojection methods",
        `Appl. Opt. 31, 1146–1152 (1992)
        <https://doi.org/10.1364/AO.31.001146>`__.

    ``two_point`` *
        Another Dasch method. Simple, and fast, but not as accurate as the
        other methods.

    The following class attributes are available, depending on the calculation.

    Returns
    -------
    transform : numpy 2D array
        the 2D forward/inverse Abel-transformed image.

    angular_integration : tuple
        (radial-grid, radial-intensity)
        radial coordinates and the radial intensity (speed) distribution,
        evaluated using :func:`abel.tools.vmi.angular_integration_3D()`.

    residual : numpy 2D array
        residual image (not currently implemented).

    IM : numpy 2D array
        the input image, re-centered (optional) with an odd-size width.

    method : str
        transform method, as specified by the input option.

    direction : str
        transform direction, as specified by the input option.

    Beta : numpy 2D array
        with ``method=linbasex, transform_options=dict(return_Beta=True)``:
        Beta array coefficients of Newton-sphere spherical harmonics

            Beta[0] - the radial intensity variation

            Beta[1] - the anisotropy parameter variation

            ...Beta[n] - higher-order terms up to ``legedre_orders=[0, ..., n]``

    radial : numpy 1D array
        with ``method=linbasex, transform_options=dict(return_Beta=True)``:
        radial grid for Beta array

    projection : numpy 2D array
        with ``method=linbasex, transform_options=dict(return_Beta=True)``:
        radial projection profiles at angles **proj_angles**

    distr : Distributions.Results object
        with ``method=rbasex``: the object from which various radial
        distributions can be retrieved
    """

    _verbose = False

    def __init__(self, IM,
                 direction='inverse', method='three_point', origin='none',
                 symmetry_axis=None, use_quadrants=(True, True, True, True),
                 symmetrize_method='average', angular_integration=False,
                 transform_options=dict(), center_options=dict(),
                 angular_integration_options=dict(),
                 recast_as_float64=True, verbose=False, center=_deprecated):
        """
        The one-stop transform function.
        """
        if center is not _deprecated:
            _deprecate('abel.transform.Transform() '
                       'argument "center" is deprecated, use "origin" instead.')
            origin = center

        # public class variables
        self.IM = IM  # (optionally) centered, odd-width image
        self.method = method
        self.direction = direction

        # private internal variables
        self._origin = origin
        self._symmetry_axis = symmetry_axis
        self._symmetrize_method = symmetrize_method
        self._use_quadrants = use_quadrants
        self._transform_options = transform_options
        self._recast_as_float64 = recast_as_float64
        _verbose = verbose

        # image processing
        self._verify_some_inputs()

        self._center_image(origin, **center_options)

        self._abel_transform_image(**transform_options)

        self._integration(angular_integration, transform_options,
                          **angular_integration_options)

    # end of class instance

    _verboseprint = print if _verbose else lambda *a, **k: None

    def _verify_some_inputs(self):
        if self.IM.ndim == 1 or np.shape(self.IM)[0] <= 2:
            raise ValueError('Data must be 2-dimensional. '
                             'To transform a single row, '
                             'use the individual transform function.')

        if not np.any(self._use_quadrants):
            raise ValueError('No image quadrants selected to use')

        if not isinstance(self._symmetry_axis, (list, tuple)):
            # if the user supplies an int, make it into a 1-element list:
            self._symmetry_axis = [self._symmetry_axis]
        elif len(self._symmetry_axis) == 0:
            # treat symmetry_axis=[] as symmetry_axis=None
            self._symmetry_axis = [None]

        if self.method == 'rbasex' and self._origin != 'none':
            if self._transform_options.get('origin') is not None:
                raise ValueError('Either use the "origin" argument to center '
                                 'the image, or pass "origin" to rbasex in '
                                 '"transform_options" to use the image as '
                                 'is, but don\'t do both.')
            if self._transform_options.get('weights') is not None:
                raise ValueError('Using the "origin" argument will center '
                                 'the image but not the "weights" array '
                                 'passed to rbasex. If you want to specify '
                                 'the image origin, pass it in '
                                 '"transform_options".')

        if self._recast_as_float64:
            self.IM = self.IM.astype('float64')

    def _center_image(self, method, **center_options):
        if method != "none":
            self.IM = tools.center.center_image(self.IM, method,
                                                **center_options)

    def _abel_transform_image(self, **transform_options):
        self._verboseprint('Calculating {0} Abel transform using {1} method -'
                          .format(self.direction, self.method),
                          '\n    image size: {:d}x{:d}'.format(*self.IM.shape))
        t0 = time.time()

        if self.method == "linbasex" and self._symmetry_axis is not None:
            self._abel_transform_image_full_linbasex(**transform_options)
        elif self.method == "rbasex":
            self._abel_transform_image_full_rbasex(**transform_options)
        else:
            self._abel_transform_image_by_quadrant(**transform_options)

        self._verboseprint("{:.2f} seconds".format(time.time() - t0))

    def _abel_transform_image_full_linbasex(self, **transform_options):
        self.transform, self.radial, self.Beta, self.projection = \
            linbasex.linbasex_transform_full(self.IM, **transform_options)

    def _abel_transform_image_full_rbasex(self, **transform_options):
        self.transform, self.distr = \
            rbasex.rbasex_transform(self.IM, direction=self.direction,
                                    **transform_options)

    def _abel_transform_image_by_quadrant(self, **transform_options):

        abel_transform = {
            "basex": basex.basex_transform,
            "daun": daun.daun_transform,
            "direct": direct.direct_transform,
            "hansenlaw": hansenlaw.hansenlaw_transform,
            "linbasex": linbasex.linbasex_transform,
            "onion_bordas": onion_bordas.onion_bordas_transform,
            "onion_peeling": dasch.onion_peeling_transform,
            "two_point": dasch.two_point_transform,
            "three_point": dasch.three_point_transform,
        }

        self._verboseprint('Calculating {0} Abel transform using {1} method -'
                          .format(self.direction, self.method),
                          '\n    image size: {:d}x{:d}'.format(*self.IM.shape))

        t0 = time.time()

        # split image into quadrants
        Q0, Q1, Q2, Q3 = tools.symmetry.get_image_quadrants(
                         self.IM, reorient=True,
                         use_quadrants=self._use_quadrants,
                         symmetry_axis=self._symmetry_axis,
                         symmetrize_method=self._symmetrize_method)

        def selected_transform(Z):
            return abel_transform[self.method](Z, direction=self.direction,
                                               **transform_options)

        AQ0 = AQ1 = AQ2 = AQ3 = None

        # Inverse Abel transform for quadrant 1 (all include Q1)
        AQ1 = selected_transform(Q1)

        if 1 not in self._symmetry_axis:
            AQ2 = selected_transform(Q2)

        if 0 not in self._symmetry_axis:
            AQ0 = selected_transform(Q0)

        if None in self._symmetry_axis:
            AQ3 = selected_transform(Q3)

        if self.method == "linbasex" and\
           "return_Beta" in transform_options.keys():
            # linbasex evaluates speed and anisotropy parameters
            # AQi == AIM, R, Beta, QLz
            Beta0 = AQ0[2]
            Beta1 = AQ1[2]
            Beta2 = AQ2[2]
            Beta3 = AQ3[2]
            # rconstructed images of each quadrant
            AQ0 = AQ0[0]
            AQ1 = AQ1[0]
            AQ2 = AQ2[0]
            AQ3 = AQ3[0]
            # speed
            self.linbasex_angular_integration = self.Beta[0]\
                 (Beta0[0] + Beta1[0] + Beta2[0] + Beta3[0])/4
            # anisotropy
            self.linbasex_anisotropy_parameter = self.Beta[1]\
                 (Beta0[1] + Beta1[1] + Beta2[1] + Beta3[1])/4

        # reassemble image
        self.transform = tools.symmetry.put_image_quadrants(
                                (AQ0, AQ1, AQ2, AQ3),
                                original_image_shape=self.IM.shape,
                                symmetry_axis=self._symmetry_axis)

        self._verboseprint("{:.2f} seconds".format(time.time()-t0))

    def _integration(self, angular_integration, transform_options,
                     **angular_integration_options):
        if angular_integration:
            if 'dr' in transform_options and\
               'dr' not in angular_integration_options:
                # assume user forgot to pass grid size
                angular_integration_options['dr'] = transform_options['dr']

            self.angular_integration = tools.vmi.angular_integration_3D(
                                             self.transform,
                                             **angular_integration_options)


# Default directory for cached basis sets;
# used by set_basis_dir() and get_basis_dir().
# DON'T access this variable directly!
_basis_dir = ''


def set_basis_dir(basis_dir='', make=True):
    """
    Changes the path to the directory for saving/loading cached basis sets that
    transform methods use by default.

    Parameters
    ----------
    basis_dir : str or None
        absolute or relative path. Passing ``''`` (default) resets to the
        system-dependent default path, see :func:`default_basis_dir`. For the
        current working directory (as in PyAbel up to v0.8.4), use ``'.'``. To
        disable basis-set caching on disk, use ``None``.
    make : bool
        create the directory if it does not exist (default: yes)

    Returns
    -------
    None
    """
    global _basis_dir

    if basis_dir == '':
        _basis_dir = default_basis_dir()
    else:
        _basis_dir = basis_dir

    if make:
        _make_basis_dir()  # (safe for None)


def get_basis_dir(make=False):
    """
    Gets the path to the directory for saving/loading cached basis sets that
    transform methods use by default. If not changed by :func:`set_basis_dir`,
    it depends on the operating system, see :func:`default_basis_dir`.

    Parameters
    ----------
    make : bool
        create the directory if it does not exist (default: no)

    Returns
    -------
    path : str or None
        absolute or relative path if disk caching is enabled,
        otherwise ``None``
    """
    global _basis_dir

    if _basis_dir == '':
        set_basis_dir('', make)

    return _basis_dir


def _make_basis_dir():
    """
    Internal utility function.

    Makes sure that the default basis-set directory (if enabled) exists.
    """
    if _basis_dir is None or os.path.exists(_basis_dir):
        return

    try:
        os.makedirs(_basis_dir)
    except Exception as e:
        print('Cannot create the directory\n"{}"\n'
              'for saving/loading basis sets:'.format(_basis_dir))
        raise


def default_basis_dir():
    r"""
    Gets full path to the system-dependent default directory for saving/loading
    cached basis sets:

    Linux (and other Unix-like):
        ``~/.cache/PyAbel`` (or ``$XDG_CACHE_HOME/PyAbel`` if set)
    macOS:
        ``/Users/<user>/Library/Caches/PyAbel``
    Windows:
        ``<user profile>\AppData\Local\PyAbel\cache`` (or
        ``%LOCALAPPDATA%\PyAbel\cache`` if set). See important notes below.

    Parameters
    ----------
    None

    Returns
    -------
    path : str
        full path to the system-dependent default basis-sets directory


    .. admonition:: Notes for MS Windows users

      * Python installed from Microsoft Store redirects subdirectory creation
        in ``AppData\Local`` to a `"private per-user, per-app location"
        <https://docs.microsoft.com/en-us/windows/msix/desktop/desktop-to-uwp-behind-the-scenes>`__
        ``AppData\Local\Packages\Python...\LocalCache\Local`` (see `Using
        Python on Windows / Known Issues
        <https://docs.python.org/3/using/windows.html#known-issues>`__).
        However, if ``AppData\Local\PyAbel\`` already exists (for example, was
        manually created *not* from Python), apparently it should be usable.

      * Old Windows versions (2000, XP, Server 2003) by default don't set the
        ``LOCALAPPDATA`` environment variable, so PyAbel will create and use
        the ``AppData\Local`` subtree in the user profile folder. This is
        probably fine, but not how it should be. To use the standard location,
        please do ::

            set LOCALAPPDATA=%USERPROFILE%\Local Settings\Application Data

        before starting Python. Or permanently set it in “Environment
        Variables” from Windows “System Properties”.
    """
    # Based on the appdirs module, https://github.com/ActiveState/appdirs
    # (see https://github.com/ActiveState/appdirs/blob/master/appdirs.py)
    system = platform.system()

    if system == 'Darwin':  # macOS
        return os.path.expanduser('~/Library/Caches/PyAbel')

    if system == 'Windows':
        return os.path.join(os.getenv('LOCALAPPDATA',
                                      os.path.expanduser(r'~\AppData\Local')),
                            r'PyAbel\cache')

    # Linux and other
    return os.path.join(os.getenv('XDG_CACHE_HOME',
                                  os.path.expanduser('~/.cache')),
                        'PyAbel')

    # system == 'Java' is ignored as useless -- Jython does not support NumPy


def basis_dir_cleanup(basis_dir='', method=None):
    """
    Deletes saved basis sets.

    Parameters
    ----------
    basis_dir : str or None
        path to the directory with saved basis sets. Use ``''`` for the default
        directory, see :func:`get_basis_dir`. (For convenience, ``None`` can be
        passed to do nothing.)
    method : str or list of str or None
        transform methods for which basis sets should be deleted. Can be a
        single string (see the :attr:`method` parameter in :class:`Transform`)
        or a list of strings. Use ``'all'`` to delete basis sets for all
        methods. ``None`` does nothing.

    Returns
    -------
    None
    """
    if basis_dir is None or method is None:
        return

    # make the list of methods
    if method == 'all':
        methods = ['basex', 'daun', 'linbasex', 'onion_peeling', 'rbasex',
                   'three_point', 'two_point']
    elif np.ndim(method) == 0:  # single string
        methods = [method]
    else:  # already a list
        methods = method

    for method in methods:
        if method in ['onion_peeling', 'three_point', 'two_point']:
            dasch.basis_dir_cleanup(method, basis_dir)
        else:
            module = sys.modules.get('abel.' + method)
            if not module:
                warn('Unknown method "{}"!'.format(method),
                     SyntaxWarning, stacklevel=2)
                continue
            func = getattr(module, 'basis_dir_cleanup', None)
            if func:
                func(basis_dir)
            else:
                warn('Method "{}" does not save basis sets.'.format(method),
                     SyntaxWarning, stacklevel=2)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from abel.tools.polar import reproject_image_into_polar
from scipy.ndimage import map_coordinates, uniform_filter1d
from scipy.ndimage.interpolation import shift
from scipy.optimize import curve_fit
from scipy.linalg import hankel, inv, pascal
from scipy.special import legendre

from abel import _deprecate


def radial_intensity(kind, IM, origin=None, dr=1, dt=None):
    """
    Calculate the one-dimensional radial intensity profile by angular
    integration or averaging of the image, treated either as a two-dimensional
    distribution or as a central slice of a cylindrically symmetric
    three-dimensional distribution.

    Parameters
    ----------
    kind : str
        operation to perform:

        ``'int2D'``:
            integration in 2D over polar angles
        ``'int3D'``:
            integration in 3D over solid angles
        ``'avg2D'``:
            averaging in 2D over polar angles
        ``'avg3D'``:
            averaging in 3D over solid angles

    IM : 2D numpy.array
        the image data

    origin : tuple of float or None
        image origin in the (row, column) format. If ``None``, the geometric
        center of the image (``rows // 2, cols // 2``) is used.

    dr : float
        radial grid spacing in pixels (default 1). ``dr=0.5`` may reduce pixel
        granularity of the radial profile.

    dt : float or None
        angular grid spacing in radians.
        If ``None``, the number of theta values will be set to largest
        dimension (the height or the width) of the image, which should
        typically ensure good sampling.

    Returns
    -------
    r : 1D numpy.array
        radial coordinates

    intensity : 1D numpy.array
        intensity profile as a function of the radial coordinate
    """
    polarIM, R, T = reproject_image_into_polar(IM, origin, dr=dr, dt=dt)

    # apply necessary Jacobian/normalization
    if kind == 'int2D':
        polarIM *= R
    elif kind == 'int3D':
        polarIM *= np.pi * R**2 * np.abs(np.sin(T))
    elif kind == 'avg2D':
        polarIM /= 2 * np.pi
    elif kind == 'avg3D':
        polarIM *= np.abs(np.sin(T)) / 4
    else:
        raise ValueError('Incorrect kind={}'.format(kind))

    # integrate over theta
    dt = T[0, 1] - T[0, 0]  # get the actual number, if dt=None was passed
    intensity = polarIM.sum(axis=1) * dt
    # (np.trapz() doesn't know about periodic functions and thus underuses
    # boundary points; direct Riemann sum is more accurate here)

    return R[:, 0], intensity


def angular_integration_2D(IM, origin=None, dr=1, dt=None):
    """
    Angular integration of the image as a two-dimensional object.

    Equivalent to :func:`radial_intensity('int2D', IM, origin, dr, dt)
    <radial_intensity>`.
    """
    return radial_intensity('int2D', IM, origin=origin, dr=dr, dt=dt)


def angular_integration_3D(IM, origin=None, dr=1, dt=None):
    """
    Angular integration of the three-dimensional cylindrically symmetric object
    represented by the image as its central slice. When applied to the inverse
    Abel transform of a velocity-mapping image, this yields the speed
    distribution.

    Equivalent to :func:`radial_intensity('int3D', IM, origin, dr, dt)
    <radial_intensity>`.
    """
    return radial_intensity('int3D', IM, origin=origin, dr=dr, dt=dt)


def average_radial_intensity_2D(IM, origin=None, dr=1, dt=None):
    """
    Calculate the average radial intensity of the image as a two-dimensional
    object.

    Equivalent to :func:`radial_intensity('avg2D', IM, origin, dr, dt)
    <radial_intensity>`.
    """
    return radial_intensity('avg2D', IM, origin=origin, dr=dr, dt=dt)


def average_radial_intensity_3D(IM, origin=None, dr=1, dt=None):
    """
    Calculate the average radial intensity of the three-dimensional
    cylindrically symmetric object represented by the image as its central
    slice.

    Equivalent to :func:`radial_intensity('avg3D', IM, origin, dr, dt)
    <radial_intensity>`.
    """
    return radial_intensity('avg3D', IM, origin=origin, dr=dr, dt=dt)


def angular_integration(IM, origin=None, Jacobian=True, dr=1, dt=None):
    r"""Angular integration of the image.

    Returns the one-dimensional intensity profile as a function of the
    radial coordinate.

    Note: the use of ``Jacobian=True`` applies the correct Jacobian for the
    integration of a 3D object in spherical coordinates.

    .. warning::
        This function behaves incorrectly: misses a factor of π for 3D
        integration, with ``Jacobian=True``, and for ``Jacobian=False`` returns
        the *average* (over polar angles) multiplied by 2π instead of
        integrating. It is currently deprecated and is provided only for
        backward compatibility, but will be removed in the future.

        Please use :func:`radial_intensity`, :func:`angular_integration_2D` or
        :func:`angular_integration_3D`.

    Parameters
    ----------
    IM : 2D numpy.array
        the image data

    origin : tuple or None
        image origin in the (row, column) format. If ``None``, the geometric
        center of the image (``rows // 2, cols // 2``) is used.

    Jacobian : bool
        Include :math:`r\sin\theta` in the angular sum (integration).
        Also, ``Jacobian=True`` is passed to
        :func:`abel.tools.polar.reproject_image_into_polar`,
        which includes another value of `r`, thus providing the appropriate
        total Jacobian of :math:`r^2\sin\theta`.

    dr : float
        radial grid spacing in pixels (default 1). ``dr=0.5`` may
        reduce pixel granularity of the speed profile.

    dt : float or None
        angular grid spacing in radians.
        If ``None``, the number of theta values will be set to largest
        dimension (the height or the width) of the image, which should
        typically ensure good sampling.

    Returns
    ------
    r : 1D numpy.array
        radial coordinates

    speeds : 1D numpy.array
        integrated intensity array (vs radius).

    """
    _deprecate('angular_integration() is deprecated, please see the '
               'documentation for details. '
               'Use radial_intensity(), angular_integration_2D() or '
               'angular_integration_3D() instead.')

    polarIM, R, T = reproject_image_into_polar(
        IM, origin, Jacobian=Jacobian, dr=dr, dt=dt)

    dt = T[0, 1] - T[0, 0]

    if Jacobian:  # × r sinθ
        polarIM *= R * np.abs(np.sin(T))

    speeds = np.trapz(polarIM, axis=1, dx=dt)

    n = speeds.shape[0]

    return R[:n, 0], speeds  # limit radial coordinates range to match speed


def average_radial_intensity(IM, **kwargs):
    """Calculate the average radial intensity of the image, averaged over all
    angles. This differs form :func:`abel.tools.vmi.angular_integration` only
    in that it returns the average intensity, and not the integrated intensity
    of a 3D image. It is equivalent to calling
    :func:`abel.tools.vmi.angular_integration` with
    ``Jacobian=True`` and then dividing the result by 2π.

    .. warning::
        This function is currently deprecated and is provided only for backward
        compatibility, but will be removed in the future.

        Please use :func:`radial_intensity`,
        :func:`average_radial_intensity_2D` or
        :func:`average_radial_intensity_3D`.

    Parameters
    ----------
    IM : 2D numpy.array
        the image data

    kwargs :
        additional keyword arguments to be passed to
        :func:`abel.tools.vmi.angular_integration`

    Returns
    -------
    r : 1D numpy.array
        radial coordinates

    intensity : 1D numpy.array
        intensity profile as a function of the radial coordinate
    """
    _deprecate('average_radial_intensity() is deprecated, '
               'use average_radial_intensity_2D(), '
               'average_radial_intensity_3D() or radial_intensity() instead.')

    R, intensity = angular_integration(IM, Jacobian=False, **kwargs)
    intensity /= 2 * np.pi
    return R, intensity


def radial_integration(IM, origin=None, radial_ranges=None):
    r""" Intensity variation in the angular coordinate.

    This function is the :math:`\theta`-coordinate complement to
    :func:`abel.tools.vmi.angular_integration`.

    Evaluates intensity vs angle for defined radial ranges.
    Determines the anisotropy parameter for each radial range.

    See :doc:`examples/example_O2_PES_PAD.py <example_O2_PES_PAD>`.

    Parameters
    ----------
    IM : 2D numpy.array
        the image data

    origin : tuple or None
        image origin in the (row, column) format. If ``None``, the geometric
        center of the image (``rows // 2, cols // 2``) is used.

    radial_ranges : list of tuple ranges or int step
        tuple
            integration ranges
            ``[(r0, r1), (r2, r3), ...]``
            evaluates the intensity vs angle
            for the radial ranges ``r0_r1``, ``r2_r3``, etc.

        int
            the whole radial range ``(0, step), (step, 2*step), ..``

    Returns
    -------
    Beta : array of tuples
        (beta0, error_beta_fit0), (beta1, error_beta_fit1), ...
        corresponding to the radial ranges

    Amplitude : array of tuples
        (amp0, error_amp_fit0), (amp1, error_amp_fit1), ...
        corresponding to the radial ranges

    Rmidpt : numpy float 1D array
        radial mid-point of each radial range

    Intensity_vs_theta: 2D numpy.array
        intensity vs angle distribution for each selected radial range

    theta: 1D numpy.array
        angle coordinates, referenced to vertical direction
    """
    if origin is not None and not isinstance(origin, tuple):
        _deprecate('radial_integration() has 2nd argument "origin", '
                   'use keyword argument "radial_ranges" or insert "None".')
        radial_ranges = origin
        origin = None

    polarIM, r_grid, theta_grid = reproject_image_into_polar(IM, origin)

    theta = theta_grid[0, :]  # theta coordinates
    r = r_grid[:, 0]          # radial coordinates

    if radial_ranges is None:
        radial_ranges = 1
    if isinstance(radial_ranges, int):
        rr = np.arange(0, r[-1], radial_ranges)
        # @DanHickstein clever code to map ranges
        radial_ranges = list(zip(rr[:-1], rr[1:]))

    Intensity_vs_theta = []
    radial_midpt = []
    Beta = []
    Amp = []
    for rr in radial_ranges:
        subr = np.logical_and(r >= rr[0], r <= rr[1])

        # sum intensity across radius of spectral feature
        intensity_vs_theta_at_R = np.sum(polarIM[subr], axis=0)
        Intensity_vs_theta.append(intensity_vs_theta_at_R)
        radial_midpt.append(np.mean(rr))

        beta, amp = anisotropy_parameter(theta, intensity_vs_theta_at_R)
        Beta.append(beta)
        Amp.append(amp)

    return Beta, Amp, radial_midpt, Intensity_vs_theta, theta


def anisotropy_parameter(theta, intensity, theta_ranges=None):
    r"""
    Evaluate anisotropy parameter :math:`\beta`, for :math:`I` vs
    :math:`\theta` data:

    .. math::

        I = \frac{\sigma_\text{total}}{4\pi} [ 1 + \beta P_2(\cos\theta) ],

    where :math:`P_2(x)=\frac{3x^2-1}{2}` is a 2nd-order Legendre polynomial.

    J. Cooper, R. N. Zare,
    "Angular Distribution of Photoelectrons",
    `J. Chem. Phys. 48, 942–943 (1968)
    <https://dx.doi.org/10.1063/1.1668742>`_.

    Parameters
    ----------
    theta : 1D numpy array
        angle coordinates, referenced to the vertical direction.

    intensity : 1D numpy array
        intensity variation with angle

    theta_ranges: list of tuples
        angular ranges over which to fit
        ``[(theta1, theta2), (theta3, theta4)]``.
        Allows data to be excluded from fit, default include all data.

    Returns
    -------
    beta : tuple of floats
        (anisotropy parameter, fit error)

    amplitude : tuple of floats
        (amplitude of signal, fit error)

    """
    def P2(x):  # 2nd-order Legendre polynomial
        return (3 * x * x - 1) / 2

    def PAD(theta, beta, amplitude):
        return amplitude * (1 + beta * P2(np.cos(theta)))  # Eq. (1) as above

    # angular range of data to be included in the fit
    if theta_ranges is not None:
        subtheta = np.ones(len(theta), dtype=bool)
        for rt in theta_ranges:
            subtheta = np.logical_and(
                subtheta, np.logical_and(theta >= rt[0], theta <= rt[1]))
        theta = theta[subtheta]
        intensity = intensity[subtheta]

    # fit angular intensity distribution
    try:
        popt, pcov = curve_fit(PAD, theta, intensity)
        beta, amplitude = popt
        error_beta, error_amplitude = np.sqrt(np.diag(pcov))
        # physical range
        if beta > 2 or beta < -1:
            beta, error_beta = np.nan, np.nan
    except:
        beta, error_beta = np.nan, np.nan
        amplitude, error_amplitude = np.nan, np.nan

    return (beta, error_beta), (amplitude, error_amplitude)


def toPES(radial, intensity, energy_cal_factor, per_energy_scaling=True,
          photon_energy=None, Vrep=None, zoom=1):
    r"""
    Convert speed radial coordinate into electron kinetic or electron binding
    energy.  Return the photoelectron spectrum (PES).

    This calculation uses a single scaling factor **energy_cal_factor**
    to convert the radial pixel coordinate into electron kinetic energy.

    Additional experimental parameters: **photon_energy** will give the
    energy scale as electron binding energy, in the same energy units,
    while **Vrep**, the VMI lens repeller voltage (volts), provides for a
    voltage-independent scaling factor. i.e. **energy_cal_factor** should
    remain approximately constant.

    The **energy_cal_factor** is readily determined by comparing the
    generated energy scale with published spectra. e.g. for O\ :sup:`−`
    photodetachment, the strongest fine-structure transition occurs at the
    electron affinity :math:`EA = 11\,784.676(7)` cm\ :math:`^{-1}`. Values for
    the ANU experiment are given below, see also
    :doc:`examples/example_hansenlaw.py <example_hansenlaw>`.

    Parameters
    ----------
    radial : numpy 1D array

        radial coordinates.

    intensity : numpy 1D array

        intensity values, at the radial array.

    energy_cal_factor : float

        energy calibration factor that will convert radius squared into energy.
        The units affect the units of the output. e.g. inputs in
        eV/pixel\ :sup:`2`, will give output energy units in eV.  A value of
        :math:`1.148427\times 10^{-5}` cm\ :math:`^{-1}/`\ pixel\ :sup:`2`
        applies for "examples/data/O-ANU1024.txt" (with Vrep = -98 volts).

    per_energy_scaling : bool

        sets the intensity Jacobian.
        If ``True``, the returned intensities correspond to an "intensity per
        eV" or "intensity per cm\ :sup:`−1`". If ``False``, the returned
        intensities correspond to an "intensity per pixel".

    photon_energy : None or float

        measurement photon energy. The output energy scale is then set to
        electron-binding-energy in units of **energy_cal_factor**. The
        conversion from wavelength (nm) to **photon_energy** (cm\ :sup:`−1`)
        is :math:`10^{7}/\lambda` (nm) e.g. ``1.0e7/812.51`` for
        "examples/data/O-ANU1024.txt".

    Vrep : None or float

        repeller voltage. Convenience parameter to allow the
        **energy_cal_factor** to remain constant, for different VMI lens
        repeller voltages. Defaults to `None`, in which case no extra scaling
        is applied. e.g. ``-98`` (volts), for "examples/data/O-ANU1024.txt".

    zoom : float

        additional scaling factor if the input experimental image has been
        zoomed.  Default 1.

    Returns
    -------
    eKBE : numpy 1D array of floats

        energy scale for the photoelectron spectrum in units of
        **energy_cal_factor**.  Note that the data is no longer on
        a uniform grid.

    PES : numpy 1D array of floats

        the photoelectron spectrum, scaled according to the
        **per_energy_scaling** input parameter.

    """

    if Vrep is not None:
        energy_cal_factor *= np.abs(Vrep) / zoom**2

    eKE = radial**2 * energy_cal_factor

    if photon_energy is not None:
        # electron binding energy
        eBKE = photon_energy - eKE
    else:
        eBKE = eKE

    # Jacobian correction to intensity, radius has been squared
    # We have E = c1 - c2 * r**2, where c1 and c2 are constants. To get thei
    # Jacobian, we find dE/dr = 2c2r. Since the coordinates are getting
    # stretched at high E and "squished" at low E, we know that we need to
    # divide by this factor.
    intensity[1:] /= (2 * radial[1:])  # 1: to exclude R = 0
    if per_energy_scaling:
        # intensity per unit energy
        intensity /= energy_cal_factor

    # sort into ascending order
    indx = eBKE.argsort()

    return eBKE[indx], intensity[indx]


class Distributions(object):
    r"""
    Class for calculating various radial distributions.

    Objects of this class hold the analysis parameters and cache some
    intermediate computations that do not depend on the image data. Multiple
    images can be analyzed (using the same parameters) by feeding them to the
    object::

        distr = Distributions(parameters)
        results1 = distr(image1)
        results2 = distr(image2)

    If analyses with different parameters are required, multiple objects can be
    used. For example, to analyze 4 quadrants independently::

        distr0 = Distributions('ll', ...)
        distr1 = Distributions('lr', ...)
        distr2 = Distributions('ur', ...)
        distr3 = Distributions('ul', ...)

        for image in images:
            Q0, Q1, Q2, Q3 = ...
            res0 = distr0(Q0)
            res1 = distr1(Q1)
            res2 = distr2(Q2)
            res3 = distr3(Q3)

    However, if all the quadrants have the same dimensions, it is more
    memory-efficient to flip them all to the same orientation and use a single
    object::

        distr = Distributions('ll', ...)

        for image in images:
            Q0, Q1, Q2, Q3 = ...
            res0 = distr(Q0)
            res1 = distr(Q1[:, ::-1])  # or np.fliplr
            res2 = distr(Q2[::-1, ::-1])  # or np.flip(Q2, (0, 1))
            res3 = distr(Q3[::-1, :])  # or np.flipud

    More concise function to calculate distributions for single images
    (without caching) are also available, see :func:`harmonics`, :func:`Ibeta`
    below.

    Parameters
    ----------
    origin : tuple of int or str
        origin of the radial distributions (the pole of polar coordinates)
        within the image.

        ``(int, int)``:
            explicit row and column indices
        ``str``:
            location string specifying the vertical and horizontal positions
            (in this order!) using the words from the following diagram::

                              left            center             right

                   top/upper  [0, 0]---------[0, n//2]--------[0, n-1]
                              |                                      |
                              |                                      |
                      center  [m//2, 0]    [m//2, n//2]    [m//2, n-1]
                              |                                      |
                              |                                      |
                bottom/lower  [m-1, 0]------[m-1, n//2]-----[m-1, n-1]

            The words can be abbreviated to their first letter each (such as
            ``'top left'`` → ``'tl'``, the space is then not required).

            ``'center center'``/``'cc'`` can also be shortened to
            ``'center'``/``'c'``.

            Examples:

                ``'center'`` or ``'cc'`` (default) for the full centered image

                ``'center left'``/``'cl'`` for the right image half, vertically
                centered

                ``'bottom left'``/``'bl'`` or ``'lower left'``/``'ll'`` for the
                upper-right image quadrant

    rmax : int or str
        largest radius to include in the distributions

        ``int``:
            explicit value
        ``'hor'``:
            fitting inside horizontally
        ``'ver'``:
            fitting inside vertically
        ``'HOR'``:
            touching horizontally
        ``'VER'``:
            touching vertically
        ``'min'``:
            minimum of ``'hor'`` and ``'ver'``, the largest area with 4 full
            quadrants
        ``'max'``:
            maximum of ``'hor'`` and ``'ver'``, the largest area with 2 full
            quadrants
        ``'MIN'`` (default):
            minimum of ``'HOR'`` and ``'VER'``, the largest area with 1 full
            quadrant (thus the largest with the full 90° angular range)
        ``'MAX'``:
            maximum of ``'HOR'`` and  ``'VER'``
        ``'all'``:
            covering all pixels (might have huge errors at large *r*, since the
            angular dependences must be inferred from very small available
            angular ranges)
    order : int
        highest order in the angular distributions, ≥ 0 (by default, 2).
        Requesting very high orders (≳ 15) can result in excessive noise,
        especially at small radii and for narrow peaks.
    odd : bool
        include odd angular orders. By default is ``False``, but is enabled
        automatically if **order** is odd. Notice that although odd orders can
        be extracted from the upper or lower image part alone, analyzing the
        whole image is more reliable.
    use_sin: bool
        use :math:`|\sin \theta|` weighting (enabled by default). This is the
        weight implied in spherical integration (for the total intensity, for
        example) and with respect to which the Legendre polynomials are
        orthogonal, so using it in the fitting procedure gives the most
        reasonable results even if the data deviates form the assumed angular
        behavior. It also reduces contributions from the centerline noise.
    weights : m × n numpy array, optional
        in addition to the optional :math:`|\sin \theta|` weighting (see
        **use_sin** above), use given weights for each pixel. The array shape
        must match the image shape.

        Parts of the image can be excluded from the fitting by assigning zero
        weights to their pixels.

        (Note: if ``use_sin=False``, a reference to this array is cached
        instead of its content, so if you modify the array between creating the
        object and using it, the results will be surprising. However, if
        needed, you can pass a copy as ``weights=weights.copy()``.)
    method : str
        numerical integration method used in the fitting procedure

        ``'nearest'``:
            each pixel of the image is assigned to the nearest radial bin. The
            fastest, but noisier (especially for high orders).
        ``'linear'`` (default):
            each pixel of the image is linearly distributed over the two
            adjacent radial bins. About twice slower than ``'nearest'``, but
            smoother.
        ``'remap'``:
            the image is resampled to a uniform polar grid, then polar pixels
            are summed over all angles for each radius. The smoothest, but
            significantly slower and might have problems with
            **rmax** > ``'MIN'`` and discontinuous weights.
    """
    def __init__(self, origin='center', rmax='MIN', order=2, odd=False,
                 use_sin=True, weights=None, method='linear'):
        # remember parameters
        self.origin = origin
        self.rmax_in = rmax
        if order < 0:
            raise ValueError('Incorrect order={}'.format(order))
        self.order = order
        if order == 0:
            self.odd = False  # (to eliminate additional checks)
        elif order % 2:
            self.odd = True  # enable automatically for odd orders
        else:
            self.odd = odd
        self.N = 1 + (order if self.odd else order // 2)  # angular terms
        self.use_sin = use_sin
        self.weights = weights
        if weights is None:
            self.shape = None
        else:
            self.shape = weights.shape
        self.method = method

        # whether precalculations are done
        self.ready = False

        # do precalculations if image size is known (from weights array)
        if self.shape is not None:
            self._precalc(self.shape)
        # otherwise postpone them to the first image

    # Note!
    # The following code has several expressions like
    #   A = w * A
    # instead of
    #   A *= w
    # This is intentional: these A can be aliases to or views of the original
    # image (passed by reference), and *= would modify A in place, thus
    # corrupting the image data owned by the caller.

    def _int_nearest(self, a, w=None):
        """
        Angular integration (radial binning) for 'nearest' method.

        a, w : arrays or None (their product is integrated)
        """
        # collect the product (if needed) in array a
        if a is None:
            a = w
        elif w is not None:
            a = w * a  # (not *=)
        # sum values from array a into bins given by array bin
        # (numpy.bincount() is faster than scipy.ndimage.sum())
        if a is not None:
            a = a.reshape(-1)
        # (if a is None, np.bincount assumes unit weights, as needed)
        return np.bincount(self.bin.reshape(-1), a, self.rmax + 2)[:-1]

    def _int_linear(self, wl, wu, a=None, w=None):
        """
        Angular integration (radial binning) for 'linear' method.

        wl, wu : lower- and upper-bin weights
        a, w : arrays or None (their product is integrated)
        """
        # collect the products (if needed) in wl, wu
        if a is None:
            a = w
        elif w is not None:
            a = w * a  # (not *=)
        if a is not None:
            # (not *=)
            wl = wl * a
            wu = wu * a
        # lower bins
        res = np.bincount(self.bin.reshape(-1),
                          wl.reshape(-1),
                          self.rmax + 2)[:-1]
        # upper bins
        res[1:] += np.bincount(self.bin.reshape(-1),
                               wu.reshape(-1),
                               self.rmax + 2)[:-2]
        return res

    def _int_remap(self, a, w=None):
        """
        Angular integration (radial binning) for 'remap' method.

        a, w : arrays or None (their product is integrated)
        """
        # collect the product (if needed) in array a
        if a is None:
            if w is None:
                return [float(self.ntheta)]
            a = w
        elif w is not None:
            a = w * a  # (not *=)
        # sum all angles together
        return a.sum(axis=0)

    def _precalc(self, shape):
        """
        Precalculate and cache quantities and structures that do not depend on
        the image data.

        shape : (rows, columns) tuple
        """
        if self.ready and shape == self.shape:  # already done
            return
        if self.weights is not None and shape != self.shape:
            raise ValueError('Image shape {} does not match weights shape {}'.
                             format(shape, self.shape))

        height, width = self.shape = shape

        # Determine origin [row, col].
        if np.ndim(self.origin) == 1:  # explicit numbers
            row, col = self.origin
            # wrap negative coordinates
            if row < 0:
                row += height
            if col < 0:
                col += width
        else:  # string with codes
            if len(self.origin) == 2:
                r, c = self.origin
            elif self.origin in ['c', 'center']:
                r, c = 'c', 'c'
            else:
                try:
                    # extract first letters
                    r, c = [word[0] for word in self.origin.split()]
                except ValueError:
                    raise ValueError('Incorrect origin "{}"'.
                                     format(self.origin))
            # vertical
            if   r in ('t', 'u'): row = 0
            elif r == 'c'       : row = height // 2
            elif r in ('b', 'l'): row = height - 1
            else:
                raise ValueError('Incorrect vertical position in "{}"'.
                                 format(self.origin))
            # horizontal
            if   c == 'l': col = 0
            elif c == 'c': col = width // 2
            elif c == 'r': col = width - 1
            else:
                raise ValueError('Incorrect horizontal position in "{}"'.
                                 format(self.origin))
        # from the other side
        row_ = height - 1 - row
        col_ = width - 1 - col
        # min/max spans
        ver, VER = min(row, row_), max(row, row_)
        hor, HOR = min(col, col_), max(col, col_)

        # save these values for rbasex
        self.row, self.col, self.VER, self.HOR = row, col, VER, HOR

        # Determine rmax.
        rmax_in = self.rmax_in
        if isinstance(rmax_in, int):
            rmax = rmax_in
        elif rmax_in == 'hor': rmax = hor
        elif rmax_in == 'ver': rmax = ver
        elif rmax_in == 'HOR': rmax = HOR
        elif rmax_in == 'VER': rmax = VER
        elif rmax_in == 'min': rmax = min(hor, ver)
        elif rmax_in == 'max': rmax = max(hor, ver)
        elif rmax_in == 'MIN': rmax = min(HOR, VER)
        elif rmax_in == 'MAX': rmax = max(HOR, VER)
        elif rmax_in == 'all': rmax = int(np.sqrt(HOR**2 + VER**2))
        else:
            raise ValueError('Incorrect radial range "{}"'.format(rmax_in))
        self.rmax = rmax

        # Folding to one quadrant with origin at [0, 0]
        # or to right half-plane for odd=True.
        # (Note: images with odd orders can be analyzed more efficiently if
        #        they have symmetric geometry and symmetric weights, but this
        #        special case would require more coding.)
        if self.odd:
            self.Qheight = Qheight = min(row, rmax) + 1 + min(row_, rmax)
            y0 = min(row, rmax)
        else:
            self.Qheight = Qheight = min(VER, rmax) + 1
            y0 = 0
        self.Qwidth = Qwidth = min(HOR, rmax) + 1
        if not self.odd and row in (0, height - 1) and col in (0, width - 1):
            # IM is already one quadrant, flip it to proper orientation
            # and possibly cut to rmax.
            if row == 0:
                self.flip_row = slice(0, Qheight)
            else:  # row == height - 1
                self.flip_row = slice(-1, -1 - Qheight, -1)
            if col == 0:
                self.flip_col = slice(0, Qwidth)
            else:  # col == width - 1
                self.flip_col = slice(-1, -1 - Qwidth, -1)
            self.fold = False
        elif self.odd and col in (0, width - 1):
            # IM is half-plane, flip it horizontally to proper orientation
            # and possibly cut.
            self.flip_row = slice(row - y0, row - y0 + Qheight)  # only cut
            if col == 0:
                self.flip_col = slice(0, Qwidth)
            else:  # col == width - 1
                self.flip_col = slice(-1, -1 - Qwidth, -1)
            self.fold = False
        else:
            # Define oriented source (IM) slices as
            # neg,neg | neg,pos
            # --------+--------
            # pos,neg | pos,pos
            # (pixel [row, col] belongs to pos,pos)
            # and corresponding destination (Q) slices.
            def slices(pivot, pivot_, size, positive):
                if positive:
                    n = min(pivot_ + 1, size)
                    return (slice(pivot, pivot + n),
                            slice(0, n))
                else:  # negative
                    n = min(pivot + 1, size)
                    return (slice(-1 - (pivot_ + 1), -1 - (pivot_ + n), -1),
                            slice(1, n))

            def slices_row(positive):
                return slices(row, row_, Qheight, positive)

            def slices_row_odd():
                return (slice(row - min(row, rmax),
                              row + 1 + min(row_, rmax)),
                        slice(0, Qheight))

            def slices_col(positive):
                return slices(col, col_, Qwidth, positive)

            # 2D region pairs (source, destination) for direct indexing
            self.regions = []
            if self.odd:
                for c in (False, True):
                    self.regions.append(list(zip(slices_row_odd(),
                                                 slices_col(c))))
            else:
                for r in (False, True):
                    for c in (False, True):
                        self.regions.append(list(zip(slices_row(r),
                                                     slices_col(c))))
            self.fold = True

        if self.method in ['nearest', 'linear']:
            # Quadrant coordinates.
            # x row
            x = np.arange(float(Qwidth))
            # y and y^2 columns
            y = y0 - np.arange(float(Qheight))[:, None]
            y2 = y**2
            # array of r^2
            r2 = x**2 + y2
            # array of r
            r = np.sqrt(r2)

            # Radial bins (as "indexing integers").
            if self.method == 'nearest':
                self.bin = r.round().astype(np.intp)
            else:  # 'linear'
                self.bin = r.astype(np.intp)  # round down (floor)
            self.bin[self.bin > rmax] = rmax + 1  # last bin is then discarded

            # Powers of cosine.
            # c[n] is cos^n for odd=True, but cos^2n for odd=False
            self.c = [None]  # (actually ones_like(r), but not used explicitly)
            if self.order > 0:
                if self.odd:
                    r[y0, 0] = np.inf  # (avoid division by zero)
                    self.c.append(y / r)  # cos theta
                    r[y0, 0] = 0  # (restore)
                else:
                    r2[y0, 0] = np.inf  # (avoid division by zero)
                    self.c.append(y2 / r2)  # cos^2 theta
                    # (r2 is not used any more, no need to restore)
                for n in range(2, 2 * self.N - 1):  # powers up to 2 × order
                    self.c.append(self.c[1] * self.c[n - 1])

            # Weights.
            if self.weights is None:
                if self.fold:
                    # count overlapping pixels
                    Qw = np.zeros((Qheight, Qwidth))
                    for src, dst in self.regions:
                        Qw[dst] += 1
                else:
                    Qw = None
            else:  # array
                if self.fold:
                    # sum all source regions into one quadrant
                    Qw = np.zeros((Qheight, Qwidth))
                    for src, dst in self.regions:
                        Qw[dst] += self.weights[src]
                else:
                    Qw = self.weights[self.flip_row, self.flip_col]

            if self.use_sin:
                r[y0, 0] = np.inf  # (avoid division by zero)
                self.Qsin = x / r
                self.Qsin[y0, 0] = 1  # (for consistency with cos[0, 0] = 0)
                r[y0, 0] = 0  # (restore)
                if Qw is None:
                    Qw = self.Qsin
                else:
                    Qw = self.Qsin * Qw  # (not *=)

            if self.method == 'linear':
                # weights for upper and lower bins
                self.wu = r - self.bin
                self.wl = 1 - self.wu

            # Integrals.
            if self.method == 'nearest':
                pc = [self._int_nearest(c, Qw) for c in self.c]
            else:  # 'linear'
                wu, wl = self.wu, self.wl
                pc = [self._int_linear(wl, wu, c, Qw) for c in self.c]
            pc = np.array(pc).T  # [r, n]

        elif self.method == 'remap':
            # Coordinates.
            if self.odd:
                thetamin = 0
            else:
                thetamin = np.pi / 2
            # angular step ~ 1 pixel at rmax
            self.ntheta = int(rmax * (np.pi - thetamin))
            # polar coordinates
            r = np.linspace(0, rmax, rmax + 1)
            theta = np.linspace(np.pi, thetamin, self.ntheta,
                                endpoint=False)[:, None]
            # rectangular coordinates of polar grid
            self.grid = np.array([y0 - r * np.cos(theta), r * np.sin(theta)])

            # Powers of cosine.
            # c[n] is cos^n for odd=True, but cos^2n for odd=False
            self.c = [None]  # (actually ones_like(r), but not used explicitly)
            if self.order > 0:
                if self.odd:
                    self.c.append(np.cos(theta))
                else:
                    self.c.append(np.cos(theta)**2)
                for n in range(2, 2 * self.N - 1):  # powers up to 2 × order
                    self.c.append(self.c[1] * self.c[n - 1])

            # Weights.
            if self.weights is None:
                if self.fold:
                    # count overlapping pixels
                    Qw = np.zeros((Qheight, Qwidth))
                    for src, dst in self.regions:
                        Qw[dst] += 1
                elif rmax > min(HOR, VER) or self.odd:
                    # (for odd=True: fold == False means only pi/2 data)
                    Qw = np.ones((Qheight, Qwidth))
                else:
                    Qw = None
            else:  # array
                if self.fold:
                    # sum all source regions into one quadrant
                    Qw = np.zeros((Qheight, Qwidth))
                    for src, dst in self.regions:
                        Qw[dst] += self.weights[src]
                else:
                    Qw = self.weights[self.flip_row, self.flip_col]
            if Qw is not None:
                Qw = map_coordinates(Qw, self.grid)

            if self.use_sin:
                self.Qsin = np.sin(theta)
                if Qw is None:
                    Qw = self.Qsin
                else:
                    Qw *= self.Qsin  # (here Qw is not aliased)

            # Integrals.
            pc = [self._int_remap(c, Qw) for c in self.c]
            pc = np.array(pc).T  # [r, n]

        else:
            raise ValueError('Incorrect method "{}"'.format(self.method))

        # higher cos powers are not needed any more
        self.c = self.c[:self.N]

        # Conversion matrices (integrals → coofficients).
        # Some r might have too few pixels and thus produce lower-rank
        # matrices. For them the highest-rank inverse is computed (including
        # as many lower orders as possible).

        # linalg.inv is very inefficient for small matrices (takes longer than
        # everything else), so they are inverted manually.
        # inv2 and inv3 are for 2×2 and 3×3 Hankel matrices (passed as unique
        # elements).
        def inv2(p):
            p0, p1, p2 = p
            d = p0 * p2 - p1 * p1
            if d == 0:
                C = np.zeros((2, 2))
                if p0 != 0:
                    C[0, 0] = 1 / p0
                return C
            return 1/d * np.array([[p2, -p1],
                                   [-p1, p0]])

        def inv3(p):
            p0, p1, p2, p3, p4 = p
            C00 = p2 * p4 - p3 * p3
            C01 = p2 * p3 - p1 * p4
            C02 = p1 * p3 - p2 * p2
            d = p0 * C00 + p1 * C01 + p2 * C02
            if d == 0:
                C = np.zeros((3, 3))
                C[:2, :2] = inv2(p[:3])
                return C
            C11 = p0 * p4 - p2 * p2
            C12 = p1 * p2 - p0 * p3
            C22 = p0 * p2 - p1 * p1
            return 1/d * np.array([[C00, C01, C02],
                                   [C01, C11, C12],
                                   [C02, C12, C22]])

        def invn(P):
            C = np.zeros((self.N, self.N))
            for m in range(self.N, 0, -1):
                try:
                    Pi = inv(P[:m, :m])
                    # due to numerical errors, inv() might "succeed" even for
                    # some degenerate matrices, so try to reject them manually
                    if np.max(Pi) > 1e14:  # (FP precision is only ~15 digits)
                        raise np.linalg.LinAlgError
                    C[:m, :m] = Pi  # (this is faster than np.pad)
                    return C
                except np.linalg.LinAlgError:
                    pass  # try lower rank
            # rank <= 1
            if P[0, 0] != 0:
                C[0, 0] = 1 / P[0, 0]
            return C

        if self.N == 1:
            pc[pc == 0] = np.inf  # to obtain inv([[0]]) = [[0]]
            self.C = 1 / pc[:, :, None]  # (new dimension to make matrices)
        elif self.N == 2:
            self.C = np.array([inv2(p) for p in pc])
        elif self.N == 3:
            self.C = np.array([inv3(p) for p in pc])
        else:
            self.C = np.array([invn(hankel(p[:self.N], p[self.N - 1:]))
                               for p in pc])

        # valid radii
        self.valid = (self.C[:, 0, 0] != 0)

        self.ready = True

    class Results(object):
        r"""
        Class for holding the results of image analysis.

        :meth:`Distributions.image` returns an object of this class, from which
        various distributions can be retrieved using the methods described
        below, for example::

            distr = Distributions(...)
            res = distr(IM)
            harmonics = res.harmonics()

        All distributions are returned as 2D arrays with the rows (1st index)
        corresponding to particular terms of the expansion and the columns (2nd
        index) corresponding to the radii. Odd angular terms are included only
        when they are used (**odd** = ``True`` or **order** is odd), otherwise
        there are only 1 + **order**/2 rows. The terms can be easily separated
        like ``I, beta2, beta4 = res.Ibeta()``. Python 3 users can also collect
        all :math:`\beta` parameters as ``I, *beta = res.Ibeta()`` for any
        **order**. Alternatively, transposing the results as ``Ibeta =
        res.Ibeta().T`` allows accessing all terms :math:`\big(I(r),
        \beta_2(r), \beta_4(r), \dots\big)` at particular radius *r* as
        ``Ibeta[r]``.

        Attributes
        ----------
        r : numpy array
            radii from 0 to **rmax**
        order : int
            highest order in the angular distributions
        odd : bool
            whether odd angular orders are present
        orders : list of int
            orders for all angular terms:

                [0, 2, ..., **order**] for **odd** = ``False``,

                [0, 1, 2, ..., **order**] for **odd** = ``True``

        sinpowers : list of int
            sine powers :math:`m` in the :math:`\cos^n\theta \cdot
            \sin^m\theta` terms from :meth:`cossin`; cosine powers :math:`n`
            are given by :attr:`orders` (see above)
        valid : bool array
            flags for each radius indicating whether it has valid data (radii
            that have zero weights for all pixels will have no valid data)
        """
        def __init__(self, r, cn, order, odd, valid=None):
            self.r = r
            self.cn = cn
            self.order = order
            self.odd = odd
            self.orders = list(range(0, order + 1, 1 if odd else 2))
            self.sinpowers = [(order - n) & ~1 for n in self.orders]
            if valid is None:
                self.valid = np.full_like(r, True)
            else:
                self.valid = valid

        def cos(self):
            r"""
            Radial distributions of :math:`\cos^n \theta` terms
            (0 ≤ *n* ≤ **order**).

            (You probably do not need them.)

            Returns
            -------
            cosn : (# terms) × (rmax + 1) numpy array
                radial dependences of the :math:`\cos^n \theta` terms, ordered
                from the lowest to the highest power
            """
            return self.cn

        def rcos(self):
            """
            Same as :meth:`cos`, but prepended with the radii row.
            """
            return np.vstack((self.r, self.cn))

        def cossin(self):
            r"""
            Radial distributions of
            :math:`\cos^n \theta \cdot \sin^m \theta` terms
            (*n* + *m* = **order**, and *n* + *m* = **order** − 1 for odd
            orders, with *m* always even).

            For **order** = 0:

                :math:`\cos^0 \theta` is the total intensity.

            For **order** = 1:

                :math:`\cos^0 \theta` is the total intensity,

                :math:`\cos^1 \theta` is the antisymmetric component.

            For **order** = 2

                :math:`\sin^2 \theta` corresponds to “perpendicular” (⟂)
                transitions,

                :math:`\cos^2 \theta` corresponds to “parallel” (∥)
                transitions.

            For **order** = 4

                :math:`\sin^4 \theta` corresponds to ⟂,⟂,

                :math:`\cos^2 \theta \cdot \sin^2 \theta` corresponds
                to ∥,⟂ and ⟂,∥,

                :math:`\cos^4 \theta` corresponds to ∥,∥.

            And so on.

            Notice that higher orders can represent lower orders as well:

               :math:`\sin^2 \theta + \cos^2 \theta= \cos^0 \theta
               \quad` (⟂ + ∥ = 1),

               :math:`\sin^4 \theta + \cos^2 \theta \cdot \sin^2 \theta
               = \sin^2 \theta \quad` (⟂,⟂ + ∥,⟂ = ⟂,⟂ + ⟂,∥ = ⟂),

               :math:`\cos^2 \theta \cdot \sin^2 \theta + \cos^4 \theta
               = \cos^2 \theta \quad` (∥,⟂ + ∥,∥ =  ⟂,∥ + ∥,∥ = ∥),

               and so forth.

            Returns
            -------
            cosnsinm : (# terms) × (rmax + 1) numpy array
                radial dependences of the :math:`\cos^n \theta \cdot \sin^m
                \theta` terms, ordered from lower to higher :math:`\cos \theta`
                powers
            """
            # conversion matrix (cos^k → cos^n sin^m) for even k
            CS = np.flip(pascal(1 + self.order // 2, 'upper'))
            # apply to all radii
            if self.odd:
                cs = np.empty_like(self.cn)
                # even powers
                cs[::2] = CS.dot(self.cn[::2])
                # odd powers
                if self.order % 2 == 0:  # even orders have
                    CS = CS[1:, 1:]  # one less odd term
                cs[1::2] = CS.dot(self.cn[1::2])
            else:
                cs = CS.dot(self.cn)
            return cs

        def rcossin(self):
            """
            Same as :meth:`cossin`, but prepended with the radii row.
            """
            return np.vstack((self.r, self.cossin()))

        def harmonics(self):
            r"""
            Radial distributions of spherical harmonics
            (Legendre polynomials :math:`P_n(\cos \theta)`).

            Spherical harmonics are orthogonal with respect to integration over
            the full sphere:

            .. math::
                \iint P_n P_m \,d\Omega =
                \int_0^{2\pi} \int_0^\pi P_n(\cos \theta) P_m(\cos \theta)
                    \,\sin\theta d\theta \,d\varphi = 0

            for *n* ≠ *m*; and :math:`P_0(\cos \theta)` is the spherically
            averaged intensity.

            Returns
            -------
            Pn : (# terms) × (rmax + 1) numpy array
                radial dependences of the :math:`P_n(\cos \theta)` terms
            """
            terms = self.cn.shape[0]
            # conversion matrix (cos^k → P_n)
            CH = np.zeros((terms, terms))
            for i in range(terms):
                if self.odd:
                    c = legendre(i).c[::-1]
                else:
                    c = legendre(2 * i).c[::-2]
                CH[:len(c), i] = c
            CH = inv(CH)
            # apply to all radii
            harm = CH.dot(self.cn)
            return harm

        def rharmonics(self):
            """
            Same as :meth:`harmonics`, but prepended with the radii row.
            """
            return np.vstack((self.r, self.harmonics()))

        def Ibeta(self, window=1):
            r"""
            Radial intensity and anisotropy distributions.

            A cylindrically symmetric 3D intensity distribution can be expanded
            over spherical harmonics (Legendre polynomials
            :math:`P_n(\cos \theta)`) as (including even and odd terms)

            .. math::
                I(r, \theta, \varphi) \, d\Omega =
                \frac{1}{4\pi} I(r) \big[1 + \beta_1(r) P_1(\cos \theta) +
                                         \beta_2(r) P_2(\cos \theta) +
                                         \dots\big],

            or, for distributions with top–bottom symmetry (only even terms),

            .. math::
                I(r, \theta, \varphi) \, d\Omega =
                \frac{1}{4\pi} I(r) \big[1 + \beta_2(r) P_2(\cos \theta) +
                                         \beta_4(r) P_4(\cos \theta) +
                                         \dots\big],

            where :math:`I(r)` is the “radial intensity distribution”
            integrated over the full sphere:

            .. math::
                I(r) = \int_0^{2\pi} \int_0^\pi I(r, \theta, \varphi)
                       \,r^2 \sin\theta d\theta \,d\varphi,

            and :math:`\beta_n(r)` are the dimensionless “anisotropy
            parameters” describing relative contributions of each harmonic
            order (:math:`\beta_0(r) = 1` by definition). In particular:

                :math:`\beta_2 = 2` for the :math:`\cos^2 \theta` (∥)
                angular distribution,

                :math:`\beta_2 = 0` for the isotropic distribution,

                :math:`\beta_2 = -1` for the :math:`\sin^2 \theta` (⟂)
                angular distribution.

            The radial intensity distribution alone for data with arbitrary
            angular variations can be obtained by using ``weight='sin'`` and
            ``order=0``.

            Parameters
            ----------
            window : int
                window size in pixels for radial averaging of :math:`\beta`.
                Since anisotropy parameters are non-linear, the central moving
                average is applied to the harmonics (which are linear), and
                then :math:`\beta` is calculated from them. In case of well
                separated peaks, setting **window** to the peak width will
                result in :math:`\beta` values at peak centers equal to total
                peak anisotropies (beware of the background, however).

            Returns
            -------
            Ibeta : (# terms) × (rmax + 1) numpy array
                radial intensity distribution (0-th term) and radial
                dependences of anisotropy parameters (other terms)
            """
            harm = self.harmonics()
            P0, Pn = np.vsplit(harm, [1])
            I = 4 * np.pi * self.r**2 * P0
            if window > 1:
                P0 = uniform_filter1d(P0, window, axis=1, mode='nearest')
                Pn = uniform_filter1d(Pn, window, axis=1, mode='nearest')
            beta = np.divide(Pn, P0, out=np.zeros_like(Pn), where=P0 != 0)
            return np.vstack((I, beta))

        def rIbeta(self, window=1):
            """
            Same as :meth:`Ibeta`, but prepended with the radii row.
            """
            return np.vstack((self.r, self.Ibeta(window)))

    def image(self, IM):
        """
        Analyze an image.

        This method can be also conveniently accessed by “calling” the object
        itself::

            distr = Distributions(...)
            Ibeta = distr(IM).Ibeta()

        Parameters
        ----------
        IM : m × n numpy array
            the image to analyze

        Returns
        -------
        results : Distributions.Results object
            the object with analysis results, from which various distributions
            can be retrieved, see :class:`Results`
        """
        # do precalculations (if needed)
        self._precalc(IM.shape)

        # apply weighting and folding
        if self.weights is not None:
            IM = self.weights * IM  # (not *=)

        if self.fold:
            Q = np.zeros((self.Qheight, self.Qwidth))
            for src, dst in self.regions:
                Q[dst] += IM[src]
        else:  # quadrant
            Q = IM[self.flip_row, self.flip_col]

        if self.method == 'remap':
            # resample to polar grid
            Q = map_coordinates(Q, self.grid)

        if self.use_sin:
            Q = self.Qsin * Q  # (not *=)

        # calculate integrals
        if self.method == 'nearest':
            p = [self._int_nearest(Q, c) for c in self.c]
        elif self.method == 'linear':  # 'linear'
            Ql, Qu = self.wl * Q, self.wu * Q
            p = [self._int_linear(Ql, Qu, c) for c in self.c]
        else:  # 'remap'
            p = [self._int_remap(Q, c) for c in self.c]

        # convert integrals to coefficients (I(r) = C(r)·p(r) for each r)
        I = np.einsum('jik,kj->ij', self.C, p)

        # radii
        r = np.arange(self.rmax + 1)

        return self.Results(r, I, self.order, self.odd, self.valid)

    def __call__(self, IM):
        return self.image(IM)


def harmonics(IM, origin='cc', rmax='MIN', order=2, **kwargs):
    """
    Convenience function to calculate harmonic distributions for a single
    image. Equivalent to ``Distributions(...).image(IM).harmonics()``.

    Notice that this function does not cache intermediate calculations, so
    using it to process multiple images is several times slower than through a
    :class:`Distributions` object.
    """
    return Distributions(origin, rmax, order, **kwargs).image(IM).harmonics()


def rharmonics(IM, origin='cc', rmax='MIN', order=2, **kwargs):
    """
    Same as :func:`harmonics`, but prepended with the radii row.
    """
    return Distributions(origin, rmax, order, **kwargs).image(IM).rharmonics()


def Ibeta(IM, origin='cc', rmax='MIN', order=2, window=1, **kwargs):
    """
    Convenience function to calculate radial intensity and anisotropy
    distributions for a single image. Equivalent to
    ``Distributions(...).image(IM).Ibeta(window)``.

    Notice that this function does not cache intermediate calculations, so
    using it to process multiple images is several times slower than through a
    :class:`Distributions` object.
    """
    return Distributions(origin, rmax, order, **kwargs).image(IM).Ibeta(window)


def rIbeta(IM, origin='cc', rmax='MIN', order=2, window=1, **kwargs):
    """
    Same as :func:`Ibeta`, but prepended with the radii row.
    """
    return Distributions(origin, rmax, order, **kwargs).image(IM).rIbeta(window)



# -*- coding: utf-8 -*-
import numpy as np
from numpy.testing import assert_allclose
import pytest

import abel


def test_direct_shape():
    """Ensure that abel.direct.direct_transform() returns an array of the correct shape"""

    n = 21
    x = np.ones((n, n))

    recon = abel.direct.direct_transform(x, direction='forward')
    assert recon.shape == (n, n) 

    recon = abel.direct.direct_transform(x, direction='inverse')
    assert recon.shape == (n, n)


def test_direct_zeros():
    """Test abel.direct.direct_transform() with zeros"""
    n = 64
    x = np.zeros((n,n))
    assert (abel.direct.direct_transform(x, direction='forward')==0).all()
    assert (abel.direct.direct_transform(x, direction='inverse')==0).all()


def test_direct_gaussian():
    """Check abel.direct.direct_transform() against the analytical Gaussian"""
    
    n = 501
    r_max = 100

    ref = abel.tools.analytical.GaussianAnalytical(n, r_max, symmetric=False,
                                                   sigma=20)

    # forward: 
    recon = abel.direct.direct_transform(ref.func, dr=ref.dr, direction='forward')
    ratio = abel.benchmark.absolute_ratio_benchmark(ref, recon, kind='forward')
    assert_allclose(ratio, 1.0, rtol=2e-2, atol=2e-2)
    
    # inverse:
    recon = abel.direct.direct_transform(ref.abel, dr=ref.dr, direction='inverse')
    ratio = abel.benchmark.absolute_ratio_benchmark(ref, recon, kind='inverse')
    assert_allclose(ratio, 1.0, rtol=2e-2, atol=2e-2)


@pytest.mark.skipif(not abel.direct.cython_ext,
                    reason='abel.direct C extension not installed')
def test_direct_c_python_correspondence_with_correction():
    """ Check that both the C and Python backends are identical (correction=True)"""
    N = 10
    r = 0.5 + np.arange(N).astype('float64') 
    x = 2*r.reshape((1, -1))**2
    out1 =  abel.direct._pyabel_direct_integral(x, r, 1)
    out2 =  abel.direct._cabel_direct_integral( x, r, 1)
    assert_allclose(out1, out2, rtol=1e-9, atol=1e-9)


@pytest.mark.skipif(not abel.direct.cython_ext,
                    reason='abel.direct C extension not installed')
def test_direct_c_python_correspondence():
    """ Check that both the C and Python backends are identical (correction=False)"""
    N = 10
    r = 0.5 + np.arange(N).astype('float64')
    x = 2*r.reshape((1, -1))**2
    
    out1 = abel.direct._pyabel_direct_integral(x, r, 0)
    out2 = abel.direct._cabel_direct_integral( x, r, 0)
    assert_allclose(out1, out2, rtol=1e-9, atol=1e-9)


if __name__ == "__main__":
    test_direct_shape()
    test_direct_zeros()
    test_direct_gaussian()
    test_direct_c_python_correspondence_with_correction()
    test_direct_c_python_correspondence()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import analytical
from . import center
from . import circularize
from . import io
from . import math
from . import polar
from . import symmetry
from . import polynomial
from . import transform_pairs
from . import vmi



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path

import numpy as np
from numpy.testing import assert_allclose
from scipy.linalg import inv, toeplitz

import abel
from abel.daun import _bs_daun, get_bs_cached, cache_cleanup, daun_transform
from abel.tools.polynomial import PiecewisePolynomial as PP
from abel.tools.analytical import GaussianAnalytical


DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')


def daun_bs(n, degree):
    """Reference implementation of basis-set generation"""
    r = np.arange(float(n))

    A = np.empty((n, n))
    for j in range(n):
        if degree == 0:
            p = PP(r, [(j - 1/2, j + 1/2, [1], j)])
        elif degree == 1:
            p = PP(r, [(j - 1, j, [1,  1], j),
                       (j, j + 1, [1, -1], j)])
        elif degree == 2:
            p = PP(r, [(j - 1,   j - 1/2, [0, 0,  2], j - 1),
                       (j - 1/2, j + 1/2, [1, 0, -2], j),
                       (j + 1/2, j + 1,   [0, 0,  2], j + 1)])
        else:  # degree == 3:
            p = PP(r, [(j - 1, j, [1, 0, -3, -2], j),
                       (j, j + 1, [1, 0, -3,  2], j)])
        A[j] = p.abel

    if degree == 3:
        B = np.empty((n, n))
        for j in range(n):
            B[j] = PP(r, [(j - 1, j, [0, 1,  2, 1], j),
                          (j, j + 1, [0, 1, -2, 1], j)]).abel
        C = toeplitz([4, 1] + [0] * (n - 2))
        C[1, 0] = C[-2, -1] = 0
        D = toeplitz([0, 3] + [0] * (n - 2), [0, -3] + [0] * (n - 2))
        D[1, 0] = D[-2, -1] = 0
        A += D.dot(inv(C)).dot(B)

    return A


def test_daun_bs():
    """Check basis-set generation for all degrees"""
    n = 10
    for degree in range(4):
        A = _bs_daun(n, degree)
        Aref = daun_bs(n, degree)
        assert_allclose(A, Aref, err_msg='-> degree = ' + str(degree))


def test_daun_bs_cache():
    """Check basis-set cache handling"""

    # changing size
    n1, n2 = 20, 30
    for degree in range(4):
        # forward
        cache_cleanup()
        bs2_ref = get_bs_cached(n2, degree, direction='forward')
        cache_cleanup()
        bs1_ref = get_bs_cached(n1, degree, direction='forward')
        bs2 = get_bs_cached(n2, degree, direction='forward')
        bs1 = get_bs_cached(n1, degree, direction='forward')
        assert_allclose(bs1, bs1_ref, atol=1e-15,
                        err_msg='-> forward: n<, degree = {}'.format(degree))
        assert_allclose(bs2, bs2_ref, atol=1e-15,
                        err_msg='-> forward: n>, degree = {}'.format(degree))

        # inverse
        for reg_type in [None, 'diff', 'L2', 'nonneg']:
            cache_cleanup()
            bs2_ref = get_bs_cached(n2, degree, reg_type, 1)
            cache_cleanup()
            bs1_ref = get_bs_cached(n1, degree, reg_type, 1)
            bs2 = get_bs_cached(n2, degree, reg_type, 1)
            bs1 = get_bs_cached(n1, degree, reg_type, 1)
            assert_allclose(bs1, bs1_ref, atol=1e-15,
                            err_msg='-> inverse: n<, degree = {}, reg_type = {}'.
                                    format(degree, reg_type))
            assert_allclose(bs2, bs2_ref, atol=1e-15,
                            err_msg='-> inverse: n>, degree = {}, reg_type = {}'.
                                    format(degree, reg_type))

    # changing degree
    for direction in ['forward', 'inverse']:
        cache_cleanup()
        bs0_ref = get_bs_cached(n1, 0)
        cache_cleanup()
        bs1_ref = get_bs_cached(n1, 1)
        bs0 = get_bs_cached(n1, 0)
        bs1 = get_bs_cached(n1, 1)
        assert_allclose(bs0, bs0_ref, atol=1e-15,
                        err_msg='-> {}: degree>'.
                                format(direction, degree, reg_type))
        assert_allclose(bs1, bs1_ref, atol=1e-15,
                        err_msg='-> {}: degree<'.
                                format(direction, degree, reg_type))

    # changing regularization strength
    bs_ref = []
    for reg in [0, 1, 2]:
        cache_cleanup()
        bs_ref.append(get_bs_cached(n1, strength=reg))
    bs = []
    for reg in [0, 1, 2]:
        bs.append(get_bs_cached(n1, strength=reg))
    assert_allclose(bs[0], bs_ref[0], atol=1e-15, err_msg='-> reg = 0 after 2')
    assert_allclose(bs[1], bs_ref[1], atol=1e-15, err_msg='-> reg = 1 after 0')
    assert_allclose(bs[2], bs_ref[2], atol=1e-15, err_msg='-> reg = 2 after 1')


def get_basis_file_name(n, degree):
    return os.path.join(DATA_DIR, 'daun_basis_{}_{}.npy'.format(n, degree))


def test_daun_bs_disk_cache():
    """Check basis-set disk caching"""
    n1, n2 = 10, 20

    f_n1_0 = get_basis_file_name(n1, 0)
    f_n2_0 = get_basis_file_name(n2, 0)
    f_n1_1 = get_basis_file_name(n1, 1)

    for fname in [f_n1_0, f_n2_0, f_n1_1]:
        if os.path.exists(fname):
            os.remove(fname)

    # saving
    cache_cleanup()
    bs_n1_0 = get_bs_cached(n1, 0, basis_dir=DATA_DIR)
    assert os.path.exists(f_n1_0), \
           'Basis set n = {}, degree = {} was not saved!'.format(n1, 0)

    bs_n2_0 = get_bs_cached(n2, 0, basis_dir=DATA_DIR)
    assert os.path.exists(f_n2_0), \
           'Basis set n = {}, degree = {} was not saved!'.format(n2, 0)

    bs_n1_1 = get_bs_cached(n1, 1, basis_dir=DATA_DIR)
    assert os.path.exists(f_n1_1), \
           'Basis set n = {}, degree = {} was not saved!'.format(n1, 1)

    # loading
    cache_cleanup()
    bs = get_bs_cached(n1, 0, basis_dir=DATA_DIR)
    assert_allclose(bs_n1_0, bs, err_msg='Loaded basis set ' +
                    'n = {}, degree = {} differs from saved!'.format(n1, 0))

    cache_cleanup()
    bs = get_bs_cached(n2, 0, basis_dir=DATA_DIR)
    assert_allclose(bs_n2_0, bs, err_msg='Loaded basis set ' +
                    'n = {}, degree = {} differs from saved!'.format(n2, 0))

    cache_cleanup()
    bs = get_bs_cached(n1, 1, basis_dir=DATA_DIR)
    assert_allclose(bs_n1_1, bs, err_msg='Loaded basis set ' +
                    'n = {}, degree = {} differs from saved!'.format(n1, 1))

    # crop-loading
    os.remove(f_n1_0)
    cache_cleanup()
    bs = get_bs_cached(n1, 0, basis_dir=DATA_DIR)
    assert_allclose(bs_n1_0, bs, err_msg='Loaded cropped basis set differs')

    # clean-up
    os.remove(f_n2_0)
    os.remove(f_n1_1)


def test_daun_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')
    recon = daun_transform(x, verbose=False)
    assert recon.shape == (n, n)


def test_daun_zeros():
    n = 21
    x = np.zeros((n, n), dtype='float32')
    recon = abel.daun.daun_transform(x, verbose=False)
    assert_allclose(recon, 0)


def test_daun_gaussian():
    """Check inverse Daun transform of a gaussian"""
    n = 100
    r_max = n - 1

    ref = GaussianAnalytical(n, r_max, symmetric=False, sigma=30)
    tr = np.tile(ref.abel[None, :], (n, 1))  # make a 2D array from 1D

    for degree, tol in [(0, 2e-3), (1, 3e-4), (2, 9e-4), (3, 4e-5)]:
        recon = daun_transform(tr, degree=degree, verbose=False)
        recon = recon[n // 2 + n % 2]
        assert_allclose(recon, ref.func, atol=tol,
                        err_msg='-> degree = ' + str(degree))

    for reg, tol in ([None, 2e-3],
                     [1, 1e-3],
                     [('diff', 1), 1e-3],
                     [('L2', 1), 6e-4],  # for r > 0
                     [('L2c', 1), 6e-4],
                     ['nonneg', 2e-3]):
        recon = daun_transform(tr, reg=reg, verbose=False)
        recon = recon[n // 2 + n % 2]
        skip = 1 if isinstance(reg, tuple) and reg[0] == 'L2' else 0
        assert_allclose(recon[skip:], ref.func[skip:], atol=tol,
                        err_msg='-> reg = ' + repr(reg))

    # test dr
    recon = daun_transform(tr, dr=0.5, verbose=False)[0]
    assert_allclose(recon, 2 * ref.func, atol=3e-3, err_msg='-> dr = 0.5')


def test_daun_forward_gaussian():
    """Check forward Daun transform of a gaussian"""
    n = 100
    r_max = n - 1

    ref = GaussianAnalytical(n, r_max, symmetric=False, sigma=30)
    tr = np.tile(ref.func[None, :], (n, 1))  # make a 2D array from 1D

    for degree, tol in [(0, 3e-2), (1, 7e-3), (2, 2e-2), (3, 7e-4)]:
        proj = daun_transform(tr, degree=degree, direction='forward',
                              verbose=False)
        proj = proj[n // 2 + n % 2]
        assert_allclose(proj, ref.abel, atol=tol,
                        err_msg='-> degree = ' + str(degree))

    # test dr
    proj = daun_transform(tr, dr=0.5, direction='forward', verbose=False)[0]
    assert_allclose(proj, ref.abel / 2, atol=2e-2, err_msg='-> dr = 0.5')


if __name__ == '__main__':
    test_daun_bs()
    test_daun_bs_cache()
    test_daun_bs_disk_cache()
    test_daun_shape()
    test_daun_zeros()
    test_daun_gaussian()
    test_daun_forward_gaussian()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from ._version import __version__

# Temporary tools to warn about deprecated functions and parameters.
from warnings import warn, filterwarnings
# class for documentation format
class __deprecated:
    def __repr__(self):
        return '<deprecated>'
# marker of deprecated parameters
_deprecated = __deprecated()
# print deprecation warning
def _deprecate(msg):
    warn(msg, DeprecationWarning, stacklevel=3)
    # 3rd level is from where the deprecated abel function is called:
    # level 0   level 1         level 2      level 3
    # warn() <- _deprecate() <- abel...() <- user code
# enable deprecation warnings (ignored by default since Python 2.7) for abel
filterwarnings('default', r'^abel\.', category=DeprecationWarning)

from . import basex
from . import benchmark
from . import dasch
from . import direct
from . import hansenlaw
from . import linbasex
from . import onion_bordas
from . import rbasex
from . import tools
from . import transform
from . import tests
from .transform import Transform
from .tools.center import center_image



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy.ndimage import map_coordinates
from scipy.ndimage.interpolation import shift
from scipy.optimize import curve_fit, minimize


def reproject_image_into_polar(data, origin=None, Jacobian=False,
                               dr=1, dt=None):
    """
    Reprojects a 2D numpy array (**data**) into a polar coordinate system,
    with the pole placed at **origin** and the angle measured clockwise from
    the upward direction. The resulting array has rows corresponding to the
    radial grid, and columns corresponding to the angular grid.

    Parameters
    ----------
    data : 2D np.array
        the image array
    origin : tuple or None
        (row, column) coordinates of the image origin. If ``None``, the
        geometric center of the image is used.
    Jacobian : bool
        Include `r` intensity scaling in the coordinate transform.
        This should be included to account for the changing pixel size that
        occurs during the transform.
    dr : float
        radial coordinate spacing for the grid interpolation.
        Tests show that there is not much point in going below 0.5.
    dt : float or None
        angular coordinate spacing (in radians).
        If ``None``, the number of angular grid points will be set to the
        largest dimension (the height or the width) of the image.

    Returns
    -------
    output : 2D np.array
        the polar image (r, theta)
    r_grid : 2D np.array
        meshgrid of radial coordinates
    theta_grid : 2D np.array
        meshgrid of angular coordinates

    Notes
    -----
    Adapted from:
    https://stackoverflow.com/questions/3798333/image-information-along-a-polar-coordinate-system

    """
    ny, nx = data.shape[:2]
    if origin is None:
        origin = (ny // 2, nx // 2)
    else:
        origin = list(origin)
        # wrap negative coordinates
        if origin[0] < 0:
            origin[0] += ny
        if origin[1] < 0:
            origin[1] += nx

    # Determine what the min and max r and theta coords will be...
    x, y = index_coords(data, origin=origin)  # (x,y) coordinates of each pixel
    r, theta = cart2polar(x, y)  # convert (x,y) -> (r,θ), note θ=0 is vertical

    nr = int(np.ceil((r.max() - r.min()) / dr))

    if dt is None:
        nt = max(nx, ny)
    else:
        # dt in radians
        nt = int(np.ceil((theta.max() - theta.min()) / dt))

    # Make a regular (in polar space) grid based on the min and max r & theta
    r_i = np.linspace(r.min(), r.max(), nr, endpoint=False)
    theta_i = np.linspace(theta.min(), theta.max(), nt, endpoint=False)
    theta_grid, r_grid = np.meshgrid(theta_i, r_i)

    # Convert the r and theta grids to Cartesian coordinates
    X, Y = polar2cart(r_grid, theta_grid)
    # then to a 2×n array of row and column indices for np.map_coordinates()
    rowi = (origin[0] - Y).flatten()
    coli = (X + origin[1]).flatten()
    coords = np.vstack((rowi, coli))

    # Remap with interpolation
    # (making an array of floats even if the data has an integer type)
    zi = map_coordinates(data, coords, output=float)
    output = zi.reshape((nr, nt))

    if Jacobian:
        output *= r_i[:, np.newaxis]

    return output, r_grid, theta_grid


def index_coords(data, origin=None):
    """
    Creates `x` and `y` coordinates for the indices in a numpy array, relative
    to the **origin**, with the `x` axis going to the right, and the `y` axis
    going `up`.

    Parameters
    ----------
    data : numpy array
        2D data. Only the array shape is used.
    origin : tuple or None
        (row, column). Defaults to the geometric center of the image.

    Returns
    -------
        x, y : 2D numpy arrays
    """
    ny, nx = data.shape[:2]
    if origin is None:
        origin_x, origin_y = nx // 2, ny // 2
    else:
        origin_y, origin_x = origin
        # wrap negative coordinates
        if origin_y < 0:
            origin_y += ny
        if origin_x < 0:
            origin_x += nx

    x, y = np.meshgrid(np.arange(float(nx)) - origin_x,
                       origin_y - np.arange(float(ny)))
    return x, y


def cart2polar(x, y):
    """
    Transform Cartesian coordinates to polar.

    Parameters
    ----------
    x, y : floats or arrays
        Cartesian coordinates

    Returns
    -------
    r, theta : floats or arrays
        Polar coordinates

    """
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan2(x, y)  # θ referenced to vertical
    return r, theta


def polar2cart(r, theta):
    """
    Transform polar coordinates to Cartesian.

    Parameters
    -------
    r, theta : floats or arrays
        Polar coordinates

    Returns
    ----------
    x, y : floats or arrays
        Cartesian coordinates
    """
    y = r * np.cos(theta)   # θ referenced to vertical
    x = r * np.sin(theta)
    return x, y



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from numpy.testing import assert_allclose

import abel

# Curve A, Table 2, Fig 3. Abel transform pair  Hansen&Law JOSA A2 510 (1985)
def f(r):
    return 1-2*r**2 if np.all(r) <= 0.5 else 2*(1-r)**2

def g(R):
    R2 = R**2
    alpha = np.sqrt(1-R**2)

    if np.all(R) <= 0.5:
        beta  = np.sqrt(0.25-R**2)
        return (2/3)*(2*alpha*(1+2*R2)-beta*(1+8*R2))-\
               4*R2*np.log((1+alpha)/(0.5+beta))
    else:
        return (4/3)*alpha*(1+2*R2)-4*R2*np.log((1+alpha)/R)


def test_onion_bordas_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')

    recon = abel.onion_bordas.onion_bordas_transform(x, direction='inverse')

    assert recon.shape == (n, n) 


def test_onion_bordas_zeros():
    n = 21
    x = np.zeros((n, n), dtype='float32')

    recon = abel.onion_bordas.onion_bordas_transform(x, direction="inverse")

    assert_allclose(recon, 0)


def test_onion_bordas_inverse_transform_gaussian():
    """Check onion_bordas inverse transform with a Gaussian function"""
    n = 501   
    r_max = 251

    ref = abel.tools.analytical.GaussianAnalytical(n, r_max, 
          symmetric=False,  sigma=10)
    tr = np.tile(ref.abel[None, :], (n, 1)) # make a 2D array from 1D

    recon = abel.onion_bordas.onion_bordas_transform(tr, ref.dr, direction='inverse')
    recon1d = recon[n//2 + n%2]  # centre row

    ratio = abel.benchmark.absolute_ratio_benchmark(ref, recon1d)

    assert_allclose(ratio, 1.0, rtol=0, atol=0.1)

def test_onion_bordas_inverse_transform_curveA():
    """ Check onion_bordas inverse transform() 'curve A'
    """
    delta = 0.01 

    # split r-domain to suit function pair
    rl = np.arange(0, 0.5+delta/2, delta) # 0 <= r <= 0.5
    rr = np.arange(0.5+delta, 1.0, delta) # 0.5 < r < 1.0
    r  = np.concatenate((rl,rr), axis=0) # whole r = [0,1)

    orig = np.concatenate((f(rl),f(rr)), axis=0)   # f(r)
    proj = np.concatenate((g(rl),g(rr)), axis=0)   # g(r)

    # inverse Abel 
    recon = abel.onion_bordas.onion_bordas_transform(proj, r[1]-r[0],
                                               direction='inverse') 
                                                       # == f(r)
    ratio = orig[20]/recon[20]

    assert_allclose(orig, recon*ratio, rtol=0, atol=0.1)

def test_onion_bordas_1d_gaussian(n=100):
    gauss = lambda r, r0, sigma: np.exp(-(r-r0)**2/sigma**2)

    n = 100
    rows, cols = n, n
    r2 = rows//2 + rows % 2
    c2 = cols//2 + cols % 2

    sigma = 20*n/100

    # 1D Gaussian -----------
    r = np.linspace(0, c2-1, c2)

    orig = gauss(r, 0, sigma)
    orig_copy = orig.copy()

    recon = abel.onion_bordas.onion_bordas_transform(orig, shift_grid=False)

    ratio_1d = np.sqrt(np.pi)*sigma

    assert_allclose(orig_copy[20:], recon[20:]*ratio_1d, rtol=0.0, atol=0.5)

def test_onion_bordas_2d_gaussian(n=100):
    gauss = lambda r, r0, sigma: np.exp(-(r-r0)**2/sigma**2)

    image_shape=(n, n)
    rows, cols = image_shape
    r2 = rows//2 + rows % 2
    c2 = cols//2 + cols % 2
    sigma = 20*n/100

    x = np.linspace(-c2, c2, cols)
    y = np.linspace(-r2, r2, rows)

    X, Y = np.meshgrid(x, y)

    IM = gauss(X, 0, sigma) # cylindrical Gaussian located at pixel R=0
    Q0 = IM[:r2, c2:] # quadrant, top-right
    Q0_copy = Q0.copy()

    # onion_bordas inverse Abel transform
    AQ0 = abel.onion_bordas.onion_bordas_transform(Q0, shift_grid=False)
    profQ0 = Q0_copy[-10:,:].sum(axis=0)
    profAQ0 = AQ0[-10:,:].sum(axis=0)

    ratio_2d = np.sqrt(np.pi)*sigma

    assert_allclose(Q0_copy, AQ0*ratio_2d, rtol=0.0, atol=0.3)

if __name__ == "__main__":
    test_onion_bordas_shape()
    test_onion_bordas_zeros()
    test_onion_bordas_1d_gaussian()
    test_onion_bordas_inverse_transform_gaussian()
    test_onion_bordas_inverse_transform_curveA()
    test_onion_bordas_2d_gaussian()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os.path

import numpy as np


# The following are just convenient functions for loading and saving images.
# Often you can just use plt.imread('my_file.png') to load a file.
# plt.imread also works for 16-bit tiff files.

def load_raw(filename, start=2, end=1440746, height=1038, width=1388):
    """
    This loads raw VMI images from Vrakking's "VMI_Acquire" software
    It ignores the first two values
    (which are just the dimensions of the image, and not actual data)
    and cuts off about 10 values at the end.
    It's unclear why the files are not quite the right size,
    but this seems to work.
    """

    # Load raw data
    A = np.fromfile(filename, dtype='int32', sep="")
    # Reshape into a numpy array
    return A[start:end].reshape([height, width])


def save16bitPNG(filename, data):
    """
    It's not easy to save 16-bit images in Python.
    Here is a way to save a 16-bit PNG
    Again, this is thanks to stackoverflow:
    https://stackoverflow.com/questions/25696615/\
    can-i-save-a-numpy-array-as-a-16-bit-image-using-normal-enthought-python
    This requires pyPNG
    """

    import png
    with open(filename, 'wb') as f:
        writer = png.Writer(width=data.shape[1], height=data.shape[0],
                            bitdepth=16, greyscale=True)
        data_list = data.tolist()
        writer.write(f, data_list)


def parse_matlab_basis_sets(path):
    """
    Load matlab generated basis sets files,
    The expected format for the `path` argument is a string of the form
    "some_basis_set_{}_1.bsc" where "{}" will be replaced by "" for
    the first file and "pr" for the second. Gzip compressed text files
    are accepted. For instance:

        basis1000_1.bst.gz

        basis1000pr_1.bst.gz
    """
    M = np.loadtxt(path.format('pr'))
    Mc = np.loadtxt(path.format(''))
    return M, Mc



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy.ndimage.interpolation import shift

################################################################################
#
# Onion peeling algorithm, also known as "back projection"
#
# This algorithm was adapted by Dan Hickstein from the original Matlab 
# implementation, created by Chris Rallis and Eric Wells of 
# Augustana University, and described in this paper:
#
#  https://scitation.aip.org/content/aip/journal/rsi/85/11/10.1063/1.4899267
#
# The algorithm actually originates from this 1996 RSI paper by Bordas et al:
#
#  https://scitation.aip.org/content/aip/journal/rsi/67/6/10.1063/1.1147044
#
# 2016-03-19: Jacobian intensity correction, allow negative image values
#             Steve Gibson and Dan Hickstein
# 2015-12-16: Python version of the Matlab code, by Dan Hickstein see issue #56
#
################################################################################

def _init_abel_vec(xc, yc):
    # vectorized 
    i = np.arange(xc, dtype=int)
    j = np.arange(yc, dtype=int)

    Ip1, I = np.meshgrid(i+1, i)
    Ijp1, Iip1 = np.meshgrid(i+1, i+1)

    val1 = np.arcsin(np.triu(Iip1/Ijp1)) - np.arcsin(np.triu(I/Ip1))

    Jp1, I1 = np.meshgrid(j+1, i+1)
    val2 = 1.0/I1[:]

    return val1, val2

def _init_abel(xc, yc):
    # this seems like it could be vectorized pretty easily
    val1 = np.zeros((xc, xc))
    val2 = np.zeros((xc, yc))

    for ii in range(xc):
        for jj in range(ii, xc):
            val1[ii, jj] = np.arcsin((ii+1)/(jj+1)) - \
                           np.arcsin((ii)/(jj+1))

        val2[ii, :] = 1.0/(ii+1)

    return val1, val2


def onion_bordas_transform(IM, dr=1, direction="inverse", shift_grid=True,
                           **kwargs):
    r"""Onion peeling (or back projection) inverse Abel transform.

    This algorithm was adapted by Dan Hickstein from the original Matlab 
    implementation, created by Chris Rallis and Eric Wells of 
    Augustana University, and described in

    C. E. Rallis, T. G. Burwitz, P. R. Andrews, M. Zohrabi, R. Averin, S. De,
    B. Bergues, B. Jochim, A. V. Voznyuk, N. Gregerson, B. Gaire,
    I. Znakovskaya, J. McKenna, K. D. Carnes, M. F. Kling, I. Ben-Itzhak,
    E. Wells,
    "Incorporating real time velocity map image reconstruction into closed-loop
    coherent control",
    `Rev. Sci. Instrum. 85, 113105 (2014)
    <https://doi.org/10.1063/1.4899267>`__.

    The algorithm actually originates from

    C. Bordas, F. Paulig,
    "Photoelectron imaging spectrometry: Principle and inversion method",
    `Rev. Sci. Instrum. 67, 2257–2268 (1996)
    <https://doi.org/10.1063/1.1147044>`__.

    This function operates on the "right side" of an image. i.e. it works on
    just half of a cylindrically symmetric image. Unlike the other transforms,
    the image origin should be at the left edge, not mid-pixel. This
    corresponds to an even-width full image.
                           
    However, ``shift_grid=True`` (default) provides the typical behavior,
    where the image origin corresponds to the pixel center in the 0th column.

    To perform a onion-peeling transorm on a whole image, use ::
    
        abel.Transform(image, method='onion_bordas').transform

    Parameters
    ----------
    IM : 1D or 2D numpy array
        right-side half-image (or quadrant)

    dr : float
        sampling size (=1 for pixel images), used for Jacobian scaling.
        The resulting inverse transform is simply scaled by 1/`dr`.

    direction : str
        only the inverse transform is currently implemented.
   
    shift_grid : bool
        place the image origin on the grid (left edge) by shifting the image
        1/2 pixel to the left.

    Returns
    -------
    AIM: 1D or 2D numpy array
        the inverse Abel transformed half-image 

    """

    if direction != 'inverse':
        raise ValueError('Forward "onion_bordas" transform not implemented')

    # make sure that the data is the right shape (1D must be converted to 2D):
    IM = np.atleast_2d(IM.copy())
    
    # onion-peeling uses grid rather than pixel values, 
    # odd shaped whole images require shift image (-1/2, -1/2)
    
    if shift_grid:
        IM = shift(IM, (0,-0.5), mode='nearest')

    # we would like to work from the outside to the inside of the image, 
    # so flip the image to put the "outside" at low index values.
    IM = np.fliplr(IM)

    h, w = np.shape(IM)

    # calculate val1 and val2, which are 2D arrays
    # of what appear to be scaling factors
    val1, val2 = _init_abel(w, h) 

    abel_arr = np.zeros_like(IM)
    # initialize 2D array for final transform
    rest_arr = IM
    # initialize 2D array that will be manipulated during the transform
    vect = np.zeros(h)
    # initialize a 1D array that is temporarily used to store scaling factors

    for col_index in range(1, w):
        # iterate over the columns (x-values) in the slice space

        idist = w - col_index
        # this is basically the x-coordinate
        # or "distance in i direction", I guess
        rest_col = rest_arr[:, col_index-1]
        # a 1D column. This is the piece of the onion we are on.
        normfac = 1 / val1[idist, idist]

        for i in range(idist)[::-1]:
            rest_arr[:, w-i-1] -= rest_col * normfac * val1[i, idist]

        for row_index in range(h):  # for row_index =1:ymax
            vect[row_index] = val2[idist, h-row_index-1]

        abel_arr[:, col_index] = normfac * rest_col * vect.transpose()

    abel_arr = np.c_[abel_arr[:, 1:],abel_arr[:, -1]]
    abel_arr = np.fliplr(abel_arr) # flip back

    # shift back to pixel grid
    if shift_grid:
        abel_arr = shift(abel_arr, (0, 1.0), mode='nearest')
    
    if abel_arr.shape[0] == 1:
        # flatten array
        abel_arr = abel_arr[0]
        
    return abel_arr/(2*dr)  # x1/2 for 'correct' normalization   



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import os
from glob import glob

import numpy as np
from scipy.linalg import inv, toeplitz, solve_banded, solve_triangular
from scipy.optimize import nnls

import abel


def daun_transform(data, reg=0.0, degree=0, dr=1.0, direction='inverse',
                   basis_dir=None, verbose=True):
    """
    Forward and inverse Abel transforms based on onion-peeling deconvolution
    using Tikhonov regularization described in

    K. J. Daun, K. A. Thomson, F. Liu, G. J. Smallwood,
    "Deconvolution of axisymmetric flame properties using Tikhonov
    regularization",
    `Appl. Opt. 45, 4638–4646 (2006)
    <https://doi.org/10.1364/AO.45.004638>`__.

    with additional basis-function types and regularization methods.

    This function operates on the “right side” of an image, that it, just one
    half of a cylindrically symmetric image, with the axial pixels located in
    the 0-th column.

    Parameters
    ----------
    data : m × n numpy array
        the image to be transformed.
        ``data[:, 0]`` should correspond to the central column of the image.
    reg : float or tuple or str
        regularization for the inverse transform:

        ``strength``:
            same as ``('diff', strength)``
        ``('diff', strength)``:
            Tikhonov regularization using the first-order difference operator
            (first-derivative approximation), as described in the original
            article
        ``('L2', strength)``:
            Tikhonov regularization using the :math:`L_2` norm, like in
            :ref:`BASEX`
        ``('L2c', strength)``:
            same as ``('L2', strength)``, but with an intensity correction
            applied to compensate the drop near the symmetry axis
        ``'nonneg'``:
            non-negative least-squares solution.

            `Warning: this regularization method is very slow, typically taking
            up to a minute for a megapixel image.`
    degree : int
        degree of basis-function polynomials:

        0:
            rectangular functions (step-function approximation), corresponding
            to “onion peeling” from the original article
        1:
            triangular functions (piecewise linear approximation)
        2:
            piecewise quadratic functions (smooth approximation)
        3:
            piecewise cubic functions (cubic-spline approximation)
    dr : float
        pixel size in the radial direction. This only affects the absolute
        scaling of the transformed image.
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed
    basis_dir : str, optional
        path to the directory for saving / loading the basis set (potentially
        useful only for **degree** = 3 and transform without regularization;
        time savings in other cases are small and might be negated by the
        disk-access overhead). Use ``''`` for the default directory. If
        ``None`` (default), the basis set will not be loaded from or saved to
        disk.
    verbose : bool
        determines whether progress report should be printed

    Returns
    -------
    recon : m × n numpy array
        the transformed (half) image
    """
    # make sure that the data has the right shape (1D must be converted to 2D)
    # and type:
    dim = len(data.shape)
    data = np.atleast_2d(data).astype(float)
    h, w = data.shape

    if reg in [None, 0]:
        reg_type, strength = None, 0
    elif reg == 'nonneg':
        reg_type, strength = reg, None
    elif isinstance(reg, (float, int)):
        reg_type, strength = 'diff', reg
    elif np.ndim(reg) == 0 and reg != 'nonneg':
        raise ValueError('Wrong regularization format "{}"'.format(reg))
    else:
        reg_type, strength = reg

    # load the basis sets and compute the transform matrix
    M = get_bs_cached(w, degree, reg_type, strength, direction, basis_dir,
                      verbose)

    if reg == 'nonneg':
        if verbose:
            print('Solving NNLS equations...')
            sys.stdout.flush()
        recon = np.empty_like(data)
        # transform row by row
        for i in range(h):
            if verbose:
                print('\r{}/{}'.format(1 + i, h), end='')
                sys.stdout.flush()
            recon[i] = nnls(M.T, data[i])[0]
        if verbose:
            print('\nDone!')
    else:
        # do the linear transform
        if direction == 'inverse' and strength == 0 and degree != 3:
            # (this is faster than general-purpose multiplication by inverse)
            recon = solve_triangular(M.T, data.T).T  # (here M is forward)
        else:
            recon = data.dot(M)

    # apply pixel scaling
    if dr != 1.0:
        if direction == 'forward':
            recon *= dr
        else:  # 'inverse'
            recon /= dr

    if dim == 1:
        return recon[0]
    else:
        return recon


# Caches and their parameters
_bs = None  # basis set
_bs_prm = None  # [size, degree]
_tr = None  # inverse-transform matrix
_tr_prm = None  # [size, type, strength]


def get_bs_cached(n, degree=0, reg_type='diff', strength=0,
                  direction='inverse', basis_dir=None, verbose=False):
    """
    Internal function.

    Gets the basis set and calculates the necessary transform matrix (notice
    that inverse direction with ``'nonneg'`` regularization, as well as with
    **strength** = 0 for **degree** ≠ 3, gives the forward (triangular)
    matrix, to be used in solvers).

    Parameters
    ----------
    n : int
        half-width of the image in pixels, must include the axial pixel
    degree : int
        polynomial degree for basis functions (0–3)
    reg_type: None or str
        regularization type (``None``, ``'diff'``, ``'L2'``, ``'L2c'``,
        ``'nonneg'``)
    strength : float
        Tikhonov regularization parameter (for **reg_type** = ``'diff'`` and
        ``'L2'``/``'L2c'``, ignored otherwise)
    direction : str: ``'forward'`` or ``'inverse'``
        type of Abel transform to be performed
    basis_dir : str or None
        path to the directory for saving / loading the basis set. Use ``''``
        for the default directory. If ``None``, the basis sets will not be
        loaded from or saved to disk.
    verbose : bool
        print some debug information

    Returns
    -------
    M : n × n numpy array
        matrix of the Abel transform (forward or inverse)
    """
    global _bs, _bs_prm, _tr, _tr_prm

    bs_OK = (_bs is not None and  # cached
             _bs_prm[1] == degree and  # correct degree
             (_bs_prm[0] == n if degree == 3 else _bs_prm[0] >= n))  # good size
    if not bs_OK:
        # try to load
        if basis_dir == '':
            basis_dir = abel.transform.get_basis_dir(make=True)
        _bs = _load_bs(basis_dir, n, degree, verbose)
        if _bs is None:
            # generate and cache
            _bs = _bs_daun(n, degree, verbose)
            _save_bs(basis_dir, n, degree, _bs, verbose)
            # (does nothing for basis_dir == None)
        _bs_prm = [n, degree]
        # reset cached inverse-transform matrix
        _tr = None
        _tr_prm = None

    if direction == 'forward' or reg_type == 'nonneg':
        return _bs[:n, :n]  # (cropping if too large)

    if reg_type is None:
        strength = 0

    if strength == 0:
        if _tr is None or _tr_prm[2] != 0:  # not cached or for strength != 0
            # compute transform matrix (full-sized, just in case)
            if degree == 3:
                # general-purpose inverse of non-triangular
                _tr = inv(_bs)
            else:
                # triangular matrix as is — will be used in solve_triangular,
                # which is even faster than multiplication by cached inverse
                _tr = _bs
            _tr_prm = [_bs_prm[0], reg_type, strength]
        return _tr[:n, :n]  # (cropping if too large — safe for triangular)

    if _tr_prm != [n, reg_type, strength]:
        # square of Tikhonov matrix
        if reg_type == 'diff':
            # of difference operator (approx. derivative operator)
            LTL = toeplitz([2, -1] + [0] * (n - 2))
            LTL[0, 0] = LTL[-1, -1] = 1
        else:
            # for L2 norm
            LTL = np.eye(n)

        # regularized inverse
        # (transposed compared to the Daun article, since our data are in rows)
        A = _bs[:n, :n]
        _tr = A.T.dot(inv(A.dot(A.T) + strength * LTL))
        if reg_type == 'L2c':
            # apply correction: divide by regularized inverse of
            # forward-transformed uniform distribution
            _tr /= A.sum(axis=0).dot(_tr)
        _tr_prm = [n, reg_type, strength]
    return _tr


def _load_bs(basis_dir, n, degree, verbose=False):
    """
    Internal function.

    Try to load the basis set from the best suitable file.

    Returns None on failure.
    """
    if basis_dir is None:
        return None

    file_mask = 'daun_basis_*_{}.npy'.format(degree)
    best_file = None
    best_size = np.inf
    for f in glob(os.path.join(basis_dir, file_mask)):
        size = int(f.split('_')[-2])  # (from '...basis_<size>_<degree>.npy')
        if n <= size < best_size:
            best_size = size
            best_file = f

    if best_file is None:
        return None

    if verbose:
        print('Loading basis set from', best_file)
    try:
        bs = np.load(best_file)
    except ValueError:
        print('Incompatible cached basis-set file!')
        return None

    if size > n:
        bs = bs[:n, :n]
        if verbose:
            print('(cropped to {})'.format(n))

    return bs


def _save_bs(basis_dir, n, degree, bs, verbose=False):
    """
    Internal function.

    Try to save the basis set.
    """
    if basis_dir is None:
        return

    file_name = 'daun_basis_{}_{}.npy'.format(n, degree)
    if verbose:
        print('Saving basis set to disk as', file_name)
    np.save(os.path.join(basis_dir, file_name), bs)


def _bs_daun(n, degree=0, verbose=False):
    """
    Internal function.

    Generate the projected basis set.

    Parameters
    ----------
    n : int
        half-width of the image in pixels, must include the axial pixel
    degree : int
        polynomial degree for basis functions (0–3)

    Returns
    -------
    A : n × n numpy array
        coefficient matrix (transposed projected basis set)
    """
    # pixel coordinates
    x = np.arange(float(n))
    # (and common subexpressions for projections)
    x2 = x**2
    if degree > 0:
        x2logx = x2 * np.log(x, np.zeros_like(x), where=x > 0)

    # projections (Abel transforms) of given-degree basis functions
    if degree == 0:
        def p(j):
            o = np.zeros(n)
            o[:j + 1] += np.sqrt((j + 1/2)**2 - x2[:j + 1])
            if j > 0:
                o[:j] -= np.sqrt((j - 1/2)**2 - x2[:j])
            return 2 * o
    elif degree == 1:
        def p(j):
            def P(R):
                y = np.sqrt(R**2 - x2[:R])
                return y * R - x2[:R] * np.log(y + R)
            o = np.zeros(n)
            o[:j + 1] += P(j + 1)
            o[:j] -= 2 * P(j)
            o[j] += x2logx[j]
            if j > 0:
                o[:j - 1] += P(j - 1)
                o[j - 1] -= x2logx[j - 1]
            return o
    elif degree == 2:
        def p(j):
            def P(R, a, b, c):
                x2_R = x2[:int(R + 1/2)]
                y = np.sqrt(R**2 - x2_R)
                return y * (a * 2 + b * R + c * 4/3 * (R**2 / 2 + x2_R)) + \
                       b * x2_R * np.log(y + R)
            o = np.zeros(n)
            o[:j + 1] += P(j + 1, 2 * (j + 1)**2, -4 * (j + 1), 2) - \
                         P(j + 1/2, (2 * j + 1)**2, -4 * (2 * j + 1), 4)
            if j > 0:
                o[:j] += P(j - 1/2, (2 * j - 1)**2, -4 * (2 * j - 1), 4)
                o[j] -= 4 * j * x2logx[j]
                o[:j - 1] -= P(j - 1, 2 * (j - 1)**2, -4 * (j - 1), 2)
                o[j - 1] += 4 * (j - 1) * x2logx[j - 1]
            return o
    elif degree == 3:
        def p(j):
            def P(R, a, b, c, d):
                y = np.sqrt(R**2 - x2[:R])
                return y * (a * 2 + (b + (c * 2/3 + d * R / 2) * R) * R +
                            (d * 3/4 * R + c * 4/3) * x2[:R]) + \
                       (b + d * 3/4 * x2[:R]) * x2[:R] * np.log(y + R)
            o = np.zeros(n)
            o[:j + 1] += P(j + 1,
                           -j**2 * (2 * j + 3) + 1, 6 * j * (j + 1),
                           -3 * (2 * j + 1),  2)
            o[:j] += P(j, 4 * j**3, -12 * j**2, 12 * j, -4)
            o[j] -= (6 * j * (j + 1) + 3/2 * x2[j]) * x2logx[j]
            if j > 0:
                o[:j - 1] -= P(j - 1,
                               j**2 * (2 * j - 3) + 1, -6 * j * (j - 1),
                               3 * (2 * j - 1), -2)
                o[j - 1] += 6 * (j * (j - 1) + x2[j - 1] / 4) * x2logx[j - 1]
            return o

        # derivative basis functions
        def q(j):
            def P(R, a, b, c, d):
                y = np.sqrt(R**2 - x2[:R])
                return y * (a * 2 + (b + (c * 2/3 + d * R / 2) * R) * R +
                            (d * 3/4 * R + c * 4/3) * x2[:R]) + \
                       (b + d * 3/4 * x2[:R]) * x2[:R] * np.log(y + R)
            o = np.zeros(n)
            o[:j + 1] += P(j + 1,
                           -j * (j * (j + 2) + 1), j * (3 * j + 4) + 1,
                           -3 * j - 2, 1)
            o[:j] += P(j, 4 * j**2, -8 * j, 4, 0)
            o[j] -= (j * (3 * j + 4) + 1 + 3/4 * x2[j]) * x2logx[j]
            if j > 0:
                o[:j - 1] -= P(j - 1,
                               -j * (j * (j - 2) + 1), j * (3 * j - 4) + 1,
                               -3 * j + 2, 1)
                o[j - 1] -= (j * (3 * j - 4) + 1 + 3/4 * x2[j - 1]) * \
                            x2logx[j - 1]
            return o
    else:
        raise ValueError('Wrong degree={} (must be 0, 1, 2 or 3).'.
                         format(repr(degree)))

    if verbose:
        print('Generating basis projections for '
              'n = {}, degree = {}...'.format(n, degree))

    # fill the coefficient matrix
    # (transposed compared to the Daun article, since our data are in rows)
    A = np.empty((n, n))
    for j in range(n):
        A[j] = p(j)

    if degree == 3:
        # coefficient matrix for derivative functions
        B = np.empty((n, n))
        for j in range(n):
            B[j] = q(j)
        # solve for smooth derivative and modify A accordingly
        C = solve_banded((1, 1), ([0] + [1] * (n - 2) + [0],
                                  [4] * n,
                                  [0] + [1] * (n - 2) + [0]),
                         3 * B)[1:-1, 1:-1]
        A[2:,  1:-1] += C
        A[:-2, 1:-1] -= C

    return A


def cache_cleanup(select='all'):
    """
    Utility function.

    Frees the memory caches created by :func:`get_bs_cached`.
    This is usually pointless, but might be required after working
    with very large images, if more RAM is needed for further tasks.

    Parameters
    ----------
    select : str
        selects which caches to clean:

        ``'all'`` (default)
            everything, including basis set
        ``'inverse'``
            only inverse transform

    Returns
    -------
    None
    """
    global _bs, _bs_prm, _tr, _tr_prm

    if select == 'all':
        _bs = None
        _bs_prm = None
    _tr = None
    _tr_prm = None


def basis_dir_cleanup(basis_dir=''):
    """
    Utility function.

    Deletes basis sets saved on disk.

    Parameters
    ----------
    basis_dir : str or None
        absolute or relative path to the directory with saved basis sets. Use
        ``''`` for the default directory. ``None`` does nothing.

    Returns
    -------
    None
    """
    if basis_dir == '':
        basis_dir = abel.transform.get_basis_dir(make=False)

    if basis_dir is None:
        return

    files = glob(os.path.join(basis_dir, 'daun_basis_*.npy'))
    for fname in files:
        os.remove(fname)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from numpy.testing import assert_allclose

import abel


def test_hansenlaw_shape():
    n = 21
    x = np.ones((n, n), dtype='float32')

    recon = abel.hansenlaw.hansenlaw_transform(x, direction='inverse')

    assert recon.shape == (n, n)


def test_hansenlaw_zeros():
    n = 21
    x = np.zeros((n, n), dtype='float32')

    recon = abel.hansenlaw.hansenlaw_transform(x, direction="inverse")

    assert_allclose(recon, 0)


def test_hansenlaw_forward_tansform_gaussian():
    """Check hansenlaw forward tansform with a Gaussian function"""
    n = 1001
    r_max = 501   # more points better fit

    ref = abel.tools.analytical.GaussianAnalytical(n,
                     r_max, symmetric=False,  sigma=200)

    recon = abel.hansenlaw.hansenlaw_transform(ref.func, ref.dr,
                                               direction='forward')

    ratio = abel.benchmark.absolute_ratio_benchmark(ref, recon, kind='forward')

    assert_allclose(ratio, 1.0, rtol=7e-2, atol=0)


def test_hansenlaw_inverse_transform_gaussian():
    """Check hansenlaw inverse transform with a Gaussian function"""
    n = 1001   # better with a larger number of points
    r_max = 501

    ref = abel.tools.analytical.GaussianAnalytical(n, r_max,
                     symmetric=False,  sigma=200)
    tr = np.tile(ref.abel[None, :], (n, 1))  # make a 2D array from 1D

    recon = abel.hansenlaw.hansenlaw_transform(tr, ref.dr, direction='inverse')
    recon1d = recon[n//2]  # center row

    ratio = abel.benchmark.absolute_ratio_benchmark(ref, recon1d)

    assert_allclose(ratio, 1.0, rtol=1e-1, atol=0)


def test_hansenlaw_forward_curveA():
    """ Check hansenlaw forward transform with 'curve A'
    """

    n = 101
    curveA = abel.tools.analytical.TransformPair(n, profile=3)

    # forward Abel == g(r)
    Aproj = abel.hansenlaw.hansenlaw_transform(curveA.func, curveA.dr,
                                               direction='forward')


    assert_allclose(curveA.abel, Aproj, rtol=0, atol=8.0e-2)


def test_hansenlaw_inverse_transform_curveA():
    """ Check hansenlaw inverse transform() 'curve A'
    """

    n = 101
    curveA = abel.tools.analytical.TransformPair(n, profile=3)

    # inverse Abel == f(r)
    recon = abel.hansenlaw.hansenlaw_transform(curveA.abel, curveA.dr,
                                               direction='inverse')

    assert_allclose(curveA.func[:n//2], recon[:n//2], rtol=0.09, atol=0)


def test_hansenlaw_forward_dribinski_image():
    """ Check hansenlaw forward/inverse transform
        using BASEX sample image, comparing speed distributions
    """

    # BASEX sample image
    IM = abel.tools.analytical.SampleImage(n=1001, name="dribinski").image

    # core transform(s) use top-right quadrant, Q0
    Q0, Q1, Q2, Q3 = abel.tools.symmetry.get_image_quadrants(IM)

    # forward Abel transform
    fQ0 = abel.hansenlaw.hansenlaw_transform(Q0, direction='forward')

    # inverse Abel transform
    ifQ0 = abel.hansenlaw.hansenlaw_transform(fQ0, direction='inverse')

    # speed distribution
    orig_speed, orig_radial = abel.tools.vmi.angular_integration_3D(Q0,
                              origin=(0, 0))

    speed, radial_coords = abel.tools.vmi.angular_integration_3D(ifQ0,
                           origin=(0, 0))

    orig_speed /= orig_speed[50:125].max()
    speed /= speed[50:125].max()

    assert np.allclose(orig_speed[50:125], speed[50:125], rtol=0.5, atol=0)

if __name__ == "__main__":
    test_hansenlaw_shape()
    test_hansenlaw_zeros()
    test_hansenlaw_forward_tansform_gaussian()
    test_hansenlaw_inverse_transform_gaussian()
    test_hansenlaw_forward_curveA()
    test_hansenlaw_inverse_transform_curveA()
    test_hansenlaw_forward_dribinski_image()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from numpy.testing import assert_allclose
from abel.tools.math import gradient


def test_gradient():
    x = np.random.randn(10,10)

    dx = np.gradient(x)[1]
    dx2 = gradient(x)

    assert_allclose(dx, dx2)




# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import warnings
from scipy import fftpack


def get_image_quadrants(IM, reorient=True, symmetry_axis=None,
                        use_quadrants=(True, True, True, True),
                        symmetrize_method="average"):
    """
    Given an image (m,n) return its 4 quadrants Q0, Q1, Q2, Q3
    as defined below.

    Parameters
    ----------
    IM : 2D np.array
        Image data shape (rows, cols)

    reorient : boolean
        Reorient quadrants to match the orientation of Q0 (top-right)

    symmetry_axis : int or tuple
        can have values of ``None``, ``0``, ``1``, or ``(0,1)`` and specifies
        no symmetry, vertical symmetry axis, horizontal symmetry axis, and both vertical
        and horizontal symmetry axes. Quadrants are added.
        See Note.

    use_quadrants : boolean tuple
        Include quadrant (Q0, Q1, Q2, Q3) in the symmetry combination(s)
        and final image

    symmetrize_method : str
        Method used for symmetrizing the image.

        ``average``
            Simply average the quadrants.

        ``fourier``
            Axial symmetry implies that the Fourier components of the 2-D
            projection should be real. Removing the imaginary components in
            reciprocal space leaves a symmetric projection.

            K. R. Overstreet, P. Zabawa, J. Tallant, A. Schwettmann,
            J. P. Shaffer,
            "Multiple scattering and the density distribution of a Cs MOT",
            `Optics Express 13, 9672–9682 (2005)
            <https://dx.doi.org/10.1364/OPEX.13.009672>`_.

    Returns
    -------
    Q0, Q1, Q2, Q3 : tuple of 2D np.arrays
        shape: (``rows//2+rows%2, cols//2+cols%2``)
        all oriented in the same direction as Q0 if ``reorient=True``


    Notes
    -----

    The symmetry_axis keyword averages quadrants like this::

        +--------+--------+
        | Q1   * | *   Q0 |
        |   *    |    *   |
        |  *     |     *  |               cQ1 | cQ0
        +--------o--------+ --(output) -> ----o----
        |  *     |     *  |               cQ2 | cQ3
        |   *    |    *   |
        | Q2  *  | *   Q3 |          cQi == combined quadrants
        +--------+--------+

        symmetry_axis = None - individual quadrants
        symmetry_axis = 0 (vertical) - average Q0+Q1, and Q2+Q3
        symmetry_axis = 1 (horizontal) - average Q1+Q2, and Q0+Q3
        symmetry_axis = (0, 1) (both) - combine and average all 4 quadrants


    The end results look like this::

        (0) symmetry_axis = None

            returned image   Q1 | Q0
                            ----o----
                             Q2 | Q3

        (1) symmetry_axis = 0

            Combine:  Q01 = Q0 + Q1, Q23 = Q2 + Q3
            returned image   Q01 | Q01
                            -----o-----
                             Q23 | Q23

        (2) symmetry_axis = 1

            Combine: Q12 = Q1 + Q2, Q03 = Q0 + Q3
            returned image   Q12 | Q03
                            -----o-----
                             Q12 | Q03

        (3) symmetry_axis = (0, 1)

            Combine all quadrants: Q = Q0 + Q1 + Q2 + Q3
            returned image   Q | Q
                            ---o---  all quadrants equivalent
                             Q | Q

    """
    IM = np.atleast_2d(IM)

    if not isinstance(symmetry_axis, (list, tuple)):
        # if the user supplies an int, make it into a 1-element list:
        symmetry_axis = [symmetry_axis]

    if ((symmetry_axis == [None] and (use_quadrants[0]==False
                                   or use_quadrants[1]==False
                                   or use_quadrants[2]==False
                                   or use_quadrants[3]==False)) or
        # at least one empty
        (symmetry_axis == [0] and use_quadrants[0]==False and
                                  use_quadrants[1]==False) or # top empty
        (symmetry_axis == [0] and use_quadrants[2]==False and
                                  use_quadrants[3]==False) or # bot empty
        (symmetry_axis == [1] and use_quadrants[1]==False and
                                  use_quadrants[2]==False) or # left empty
        (symmetry_axis == [1] and use_quadrants[0]==False and
                                  use_quadrants[3]==False)    # right empty
                                                                or
        not np.any(use_quadrants)
        ):
        raise ValueError('At least one quadrant would be empty.'
                         ' Please check symmetry_axis and use_quadrant'
                         ' values to ensure that all quadrants will have a'
                         ' defined value.')

    n, m = IM.shape

    # odd size increased by 1
    n_c = n // 2 + n % 2
    m_c = m // 2 + m % 2


    if isinstance(symmetry_axis, tuple) and not reorient:
        raise ValueError(
            'In order to add quadrants (i.e., to apply horizontal or \
            vertical symmetry), you must reorient the image.')

    if symmetrize_method == "fourier":
        if np.sum(use_quadrants)<4:
            warnings.warn("Using Fourier transformation to symmetrize the"
                          " data will use all 4 quadrants!!")
        if 0 in symmetry_axis:
            IM = fftpack.ifft(fftpack.fft(IM).real).real
        if 1 in symmetry_axis:
            IM = fftpack.ifft(fftpack.fft(IM.T).real).T.real

    # define 4 quadrants of the image
    # see definition above
    Q0 = IM[:n_c, -m_c:]*use_quadrants[0]
    Q1 = IM[:n_c, :m_c]*use_quadrants[1]
    Q2 = IM[-n_c:, :m_c]*use_quadrants[2]
    Q3 = IM[-n_c:, -m_c:]*use_quadrants[3]

    if reorient:
        Q1 = np.fliplr(Q1)
        Q3 = np.flipud(Q3)
        Q2 = np.fliplr(np.flipud(Q2))

    if symmetrize_method == "fourier":
        return Q0, Q1, Q2, Q3

    elif symmetrize_method == "average":
        if symmetry_axis==(0, 1):
            Q = (Q0 + Q1 + Q2 + Q3)/np.sum(use_quadrants)
            return Q, Q, Q, Q

        if 0 in symmetry_axis:   #  vertical axis image symmetry
            Q0 = Q1 = (Q0 + Q1)/(use_quadrants[0] + use_quadrants[1])
            Q2 = Q3 = (Q2 + Q3)/(use_quadrants[2] + use_quadrants[3])

        if 1 in symmetry_axis:   # horizontal axis image symmetry
            Q1 = Q2 = (Q1 + Q2)/(use_quadrants[1] + use_quadrants[2])
            Q0 = Q3 = (Q0 + Q3)/(use_quadrants[0] + use_quadrants[3])

        return Q0, Q1, Q2, Q3
    else:
        raise ValueError("Invalid method for symmetrizing the image!!")


def put_image_quadrants(Q, original_image_shape, symmetry_axis=None):
    """
    Reassemble image from 4 quadrants Q = (Q0, Q1, Q2, Q3)
    The reverse process to get_image_quadrants(reorient=True)

    Note: the quadrants should all be oriented as Q0, the upper right quadrant

    Parameters
    ----------
    Q : tuple of np.array (Q0, Q1, Q2, Q3)
        Image quadrants all oriented as Q0
        shape (``rows//2+rows%2, cols//2+cols%2``) ::

            +--------+--------+
            | Q1   * | *   Q0 |
            |   *    |    *   |
            |  *     |     *  |
            +--------o--------+
            |  *     |     *  |
            |   *    |    *   |
            | Q2  *  | *   Q3 |
            +--------+--------+

    original_image_shape : tuple
        (rows, cols)

        reverses the padding added by `get_image_quadrants()` for odd-axis sizes

        odd row trims 1 row from Q1, Q0

        odd column trims 1 column from Q1, Q2

    symmetry_axis : int or tuple
        impose image symmetry

            ``symmetry_axis = 0 (vertical)   - Q0 == Q1 and Q3 == Q2``
            ``symmetry_axis = 1 (horizontal) - Q2 == Q1 and Q3 == Q0``


    Returns
    -------
    IM : np.array
        Reassembled image of shape (rows, cols): ::

            symmetry_axis =

            None             0              1           (0,1)

             Q1 | Q0        Q1 | Q1        Q1 | Q0       Q1 | Q1
            ----o----  or  ----o----  or  ----o----  or ----o----
             Q2 | Q3        Q2 | Q2        Q1 | Q0       Q1 | Q1

    """

    Q0, Q1, Q2, Q3 = Q

    if not isinstance(symmetry_axis, (list, tuple)):
        # if the user supplies an int, make it into a 1-element list:
        symmetry_axis = [symmetry_axis]

    if 0 in symmetry_axis:
        Q0 = Q1
        Q3 = Q2

    if 1 in symmetry_axis:
        Q2 = Q1
        Q3 = Q0

    if original_image_shape[0] % 2:
        # odd-rows => remove duplicate bottom row from Q1, Q0
        Q0 = Q0[:-1, :]
        Q1 = Q1[:-1, :]

    if original_image_shape[1] % 2:
        # odd-columns => remove duplicate first column from Q1, Q2
        Q1 = Q1[:, 1:]
        Q2 = Q2[:, 1:]

    Top = np.concatenate((np.fliplr(Q1), Q0), axis=1)
    Bottom = np.flipud(np.concatenate((np.fliplr(Q2), Q3), axis=1))

    IM = np.concatenate((Top, Bottom), axis=0)

    return IM



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

import numpy as np
from numpy.testing import assert_array_equal

from abel.transform import get_basis_dir, set_basis_dir


def test_basis_dir():
    """
    Test basic functionality of basis_dir tools.
    """
    # check that default directory is not '' or None and can be created
    basis_dir = get_basis_dir(make=True)
    assert basis_dir, 'basis_dir = {}'.format(repr(basis_dir))

    # check saving to default directory
    path = os.path.join(basis_dir, 'test.npy')
    saved = np.random.randn(10, 10)
    np.save(path, saved)

    # check that default directory can be set to None
    set_basis_dir(None)
    tmp = get_basis_dir()
    assert tmp is None, \
           'basis_dir = {} after set_basis_dir(None)'.format(repr(tmp))

    # check that default directory can be reset to system default
    set_basis_dir('', make=False)
    tmp = get_basis_dir()
    assert tmp == basis_dir, \
           'basis_dir = {} after set_basis_dir("")'.format(repr(tmp))

    # check loading from default directory
    loaded = np.load(path)
    assert_array_equal(loaded, saved, err_msg='Loaded != saved')

    # remove temporary file
    os.remove(path)


if __name__ == "__main__":
    test_basis_dir()
